//-----------------------------------------------------------------------------
//
// Title       : Error_corr
// Design      : decoder_5bit
// Author      : Sindre
// File        : Error_corr.v
//
//-----------------------------------------------------------------------------
//
// Description : Error correction module. Corrects the error found by
//				 Chien search, by using the value calcualted in the Forney module
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

//{module {Error_corr}}
module Error_corr #(parameter width=5)(
	input clk,							//clock signal
	input reset,						//reset signal
	input enable,						//enable signal
	input EV_ready,						//error value ready signal
	input [width-1:0] E_value,			//error value
	input [width-1:0] S_IN,				//symbol input from codeword storage
	output reg RE,						//read signal (to read symbol values from codeword storage)
	output reg RS_DONE,					//RS decoding done
	output reg [width-1:0] add_read,	//RAM read address 
	output [width-1:0] S_OUT		//corrected symbol value
	);
	
	//internal constants
	parameter size = 3;
	parameter IDLE = 3'b000;
	parameter STEP1 = 3'b001;
	parameter STEP2 = 3'b011;
	
	//internal variables
	reg [width-1:0] value; 		//internal reg for error value
	reg [4:0] cnt;				//counter: max 31 - number of symbols
	reg [size-1:0] state;
	reg [size-1:0] next_state;
	
	assign S_OUT = S_IN^value;	//perform finite field addition
	
	//FSM
	always @(posedge clk or negedge reset)
	begin: FSM_SEQ
		if(reset == 0)begin
			state <= IDLE;
		end else if(enable)begin
			state <= next_state;	
		end
	end
	
	//state change section
	always @(state or EV_ready or cnt)
	begin: FSM_COMBO
		case(state)
			IDLE: begin
				if(EV_ready)begin
					next_state = STEP1;	
				end else begin
					next_state = IDLE;
				end
			end
			
			STEP1: begin
				if(cnt == 0)begin
					next_state = STEP2;	
				end else begin
					next_state = STEP1;
				end
			end
			
			STEP2: begin
				if(cnt == 0)begin
					next_state = IDLE;
				end else begin
					next_state = STEP2;
				end
			end
			
			default: next_state = IDLE;
		endcase
	end
	
	//output section
	always @(posedge clk or negedge reset)
	begin: FSM_OUT
		if(reset == 0)begin
		   RE <= 0;
		   cnt <= 0;
		   value <= 0;
		   RS_DONE <= 0;
		   add_read <= 0;
		end else if(enable) begin
			case(state)
				IDLE: begin
					if(EV_ready)begin
						//start reading values from memory when error
						//value is ready
						RE <= 1;
						cnt <= 29;
						add_read <= 30;
						value <= E_value;  //error value input
					end else begin
						RE <= 0;
						cnt <= 0;
						add_read <= 0;
					end
				end
				
				//go through all symbols
				//and correct errors
				STEP1: begin
					RE <= 1;
					value <= E_value; //error value input
					add_read <= cnt;
					if(cnt == 0)begin
						cnt <= 1;
					end else begin
						cnt <= cnt-1;
					end
					if(cnt == 29)begin
						//start shifting out symbols from decoder
						RS_DONE <= 1;
					end else begin
						RS_DONE <= 0;
					end
				
				end
				
				STEP2: begin
					//shift out the two last symbols
					value <= E_value;
					if(cnt == 1)begin
						cnt <= 0;
					end else begin
						cnt <= cnt-1;
					end		
				end
				
				default: begin
					RE <= 0;
		   			cnt <= 0;
		   			RS_DONE <= 0;
					add_read <= 0;
				end
			endcase
		end
	end

endmodule
