//-----------------------------------------------------------------------------
//
// Title       : Original_BM
// Design      : decoder_5bit
// Author      : Sindre
// File        : Original_BM.v
//
//-----------------------------------------------------------------------------
//
// Description : Original Berlekam-Massey algorithm
//				 Calcualtes error locator polynomial
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

//{module {Original_BM}}
module Original_BM #(parameter width=5)(
	input clk,				   //clock signal
	input reset,			   //reset signal
	input enable,			   //enable signal
	input S_ready,			   //syndromes ready signal
	input [width-1:0] S1,	   //syndrome value 1
	input [width-1:0] S2,	   //syndrome value 2
	input [width-1:0] S3,	   //syndrome value 3
	input [width-1:0] S4,	   //syndrome value 4
	input [width-1:0] INV,	   //inverse value input
	output reg L_ready,		   //error locator poly. ready signal
	output reg RE,			   //ROM read enable signal
	output reg [width-1:0] VAL,//value to find inverse of
	output reg [width-1:0] L0, //locator poly. coefficient 0 output
	output reg [width-1:0] L1, //locator poly. coefficient 1 output
	output reg [width-1:0] L2  //locator poly. coefficient 2 output
	);
	
	//internal constants
	parameter t = 2;
	parameter size = 3;
	parameter IDLE = 3'b000; 	//0
	parameter STEP1 = 3'b001;	//1
	parameter STEP2 = 3'b011;	//3
	parameter STEP3 = 3'b010;	//2
	parameter STEP4 = 3'b110;	//6
	parameter STEP5	= 3'b111;	//7
	parameter STEP6 = 3'b101;	//5
	
	//internal variables //temperary,
	reg [width-1:0] La [0:3]; 	// variable for storing error locator polynomial coeff
	reg [width-1:0] Lm [0:3];	// approximation for error locator poly
	reg [width-1:0] C [0:3]; 	// Correction poly
	reg [width-1:0] D;			// discrepancy (uppercase Delta)
	reg [width-1:0]	L;			// 
	reg [2:0] K;				//counter, max value 4 
	reg [2:0] i;				//counter, max value 2
	reg n_reset;
	
	reg [width-1:0] S [0:3];	// internal storage for syndrome values
	
	reg [size-1:0] state;
	reg [size-1:0] next_state;	
	
	//input variables to modules
	reg [width-1:0] M01;		//M<value> indicates an input to a multiplier
	reg [width-1:0] M02;
	reg [width-1:0] M03;
	reg [width-1:0] M04;
	reg [width-1:0] M06;
	reg [width-1:0] A02;	   //A<value> indicates an input to a adder
	//internal wires
	wire [width-1:0] NET01;	   //NET<value> indicates an internal wire
	wire [width-1:0] NET02;
	wire [width-1:0] SUM01;	   //SUM<value> indicates an output from an multiplier or adder
	wire [width-1:0] SUM02;
	wire [width-1:0] SUM03;
	wire [width-1:0] D_OUT01;
	
	
	//module instantiations
	//modules used for performing finite field addition and multiplication

	//calculate the discrepancy delta
	Full_mult #(.width(width)) FFM01(
	.A	(M01),
	.B	(M02),
	.Y	(NET01)
	);
	
	Adder #(.width(width)) FFA01(
	.in1	(NET01),
	.in2	(D_OUT01),
	.sum	(SUM01)
	);
	
	D_reg #(.width(width)) REG01(
	.clk	(clk),
	.reset	(n_reset),
	.enable	(enable),
	.d_in	(SUM01),
	.d_out	(D_OUT01)
	);
	
	//calculate new approximation to the error locator poly
	Full_mult #(.width(width)) FFM02(
	.A	(M03),
	.B	(M04),
	.Y	(NET02)
	);
	
	Adder #(.width(width)) FFA02(
	.in1	(NET02),
	.in2	(A02),
	.sum	(SUM02)
	);
	
	//forming a new correction poly
	Full_mult #(.width(width)) FFM03(
	.A	(INV),
	.B	(M06),
	.Y	(SUM03)
	);
	
	//FSM
	always @(posedge clk or negedge reset)
		begin: FSM_SEQ
			if (reset == 0) begin
				state <= IDLE;
			end else if(enable) begin
				state <= next_state;
			end
		end
	
	//change state section
	always @(state or S_ready or K or i or L or SUM01)
	begin: FSM_COMBO
		case(state)
			IDLE: begin
				if(S_ready)begin
						next_state = STEP1;
				end else begin
					next_state = IDLE;
				end
			end
			
			STEP1: begin
				if(i <= L)begin
				 	next_state = STEP1;
				end else begin
					if(SUM01 == 0)begin
						next_state = STEP2;				
					end else begin
						next_state = STEP3;		
					end
				end
			end
			
			STEP2: begin
				if(K == 2*t)begin
					next_state = STEP6;
				end else begin
					next_state = STEP1;
				end
			end
			
			STEP3: begin
				if(i == 4)begin
					if(2*L < K)begin
						next_state = STEP4;	
					end else begin
						next_state = STEP5;
					end
				end else begin
					next_state = STEP3;
				end
			end
			
			STEP4: begin
				if(i == 4)begin
					next_state = STEP5;	
				end else begin
					next_state = STEP4;
				end
			end
	
			STEP5: begin
				if(K == 2*t)begin
					next_state = STEP6;		
				end else begin
					next_state = STEP1;				
				end	
			end
			
			STEP6: begin
				next_state = IDLE;	
			end
			
			default: next_state = IDLE;
		endcase	
	end
	
	//output section
	always @(posedge clk or negedge reset)
		begin: FSM_OUT
			if(reset == 0)begin
				//reset all internal regs
				n_reset <= 1;
				L_ready <= 0;
				L0 <= 0;
				L1 <= 0;
				L2 <= 0;
				M01 <= 0;
				M02 <= 0;
				M03 <= 0;
				M04 <= 0;
				M06 <= 0;
				A02 <= 0;
				La[0] <= 0;
				La[1] <= 0;
				La[2] <= 0;
				La[3] <= 0;
				Lm[0] <= 0;
				Lm[1] <= 0;
				Lm[2] <= 0;
				Lm[3] <= 0;
				C[0] <= 0;
				C[1] <= 0;
				C[2] <= 0;
				C[3] <= 0;
				S[0] <= 0;
				S[1] <= 0;
				S[2] <= 0;
				S[3] <= 0;
				D <= 0;
				L <= 0;
				K <= 0;
				i <= 0;
				RE <= 0;
				VAL <= 0;
			end else if(enable) begin
				case(state)
					IDLE: begin
						L_ready <= 0;
						La[0] <= 1;
						La[1] <= 0;
						La[2] <= 0;
						La[3] <= 0;
						Lm[0] <= 0;
						Lm[1] <= 0;
						Lm[2] <= 0;
						Lm[3] <= 0;
						C[0] <= 0;
						C[1] <= 1;
						C[2] <= 0;
						C[3] <= 0;
						L <= 0;
						K <= 1;
						i <= 0;
						D <= 0;
						if(S_ready)begin
							//start calulation of poly when syndromes ready
							S[0] <= S1;
							S[1] <= S2;
							S[2] <= S3;
							S[3] <= S4;
							n_reset <= 0;
						end else begin
							S[0] <= 0; 
							S[1] <= 0;
							S[2] <= 0;
							S[3] <= 0;
							n_reset <= 1;
						end
					end
					
					STEP1: begin
						//start calculations
						n_reset <= 1;
						if(i <= L)begin
							//calcualte discrepancy
							M01 <= S[K-1-i];
							M02 <= La[i];
							i <= i+1;
							D <= 0;
						end else begin
							M01 <= 0;
							M02 <= 0;
							D <= SUM01;
							i <= 0;
						end
					end
					
					STEP2: begin
						//calcualte correctiom term
						n_reset <= 1;
						C[0] <= 0;
						C[1] <= C[0];
						C[2] <= C[1];
						C[3] <= C[2];
						if(K == 2*t)begin
							K <= K;
							n_reset <= 1;
						end else begin
							K <= K+1;
							n_reset <= 0;
						end
					end
					
					STEP3: begin
						if(i < 4)begin
							M03 <= D;
							M04 <= C[i];
							A02 <= La[i];
							i <= i+1;
						end else begin
							M03 <= 0;
							M04 <= 0;
							A02 <= 0;
							i <= 0;
							if(2*L > K-1)begin
								RE <= 0;
								VAL <= 0;
							end else begin
								//find inverse value
								RE <= 1;
								VAL <= D;
							end
						end
						// approximation for error locator poly
						if(i >= 1)begin
							Lm[i-1] <= SUM02;
						end else begin
							Lm[i] <= Lm[i];
						end
					end
					
					STEP4: begin
						//intermediate error locaror poly calcualtion
						RE <= 0;
						if(i < 4)begin
							M06 <= La[i];
							i <= i+1;
						end else begin
							M06 <= 0;
							i <= 0;
						end
						if(i >= 1)begin
							C[i-1] <= SUM03;
						end else begin
							C[i] <= C[i];
						end
						
						if(i == 4)begin
							L <= K-L;	
						end else begin
							L <= L;
						end 
					end
					
					STEP5: begin
						//intermediate error locaror poly results
						n_reset <= 0;
						La[0] <= Lm[0];
						La[1] <= Lm[1];
						La[2] <= Lm[2];
						La[3] <= Lm[3];
						C[0] <= 0;
						C[1] <= C[0];
						C[2] <= C[1];
						C[3] <= C[2];
						if(K == 2*t)begin
							n_reset <= 1;
							K <= K;
						end else begin
							n_reset <= 0;
							K <= K+1;
						end
					end
					
					STEP6: begin
						//output calcualted error locator poly.
						L_ready <= 1;
						L0 <= La[0];
						L1 <= La[1];
						L2 <= La[2];
					end
					
					default: begin
						n_reset <= 1;
						L_ready <= 0;
						L0 <= 0;
						L1 <= 0;
						L2 <= 0;
						M01 <= 0;
						M02 <= 0;
						M03 <= 0;
						M04 <= 0;
						M06 <= 0;
						A02 <= 0;
						La[0] <= 0;
						La[1] <= 0;
						La[2] <= 0;
						La[3] <= 0;
						Lm[0] <= 0;
						Lm[1] <= 0;
						Lm[2] <= 0;
						Lm[3] <= 0;
						C[0] <= 0;
						C[1] <= 0;
						C[2] <= 0;
						C[3] <= 0;
						S[0] <= 0;
						S[1] <= 0;
						S[2] <= 0;
						S[3] <= 0;
						D <= 0;
						L <= 0;
						K <= 0;
						i <= 0;
						RE <= 0;
						VAL <= 0;
					end
				endcase
			end
		end
endmodule
