//-----------------------------------------------------------------------------
//
// Title       : Syndrome_ctrl
// Design      : decoder_5bit
// Author      : Sindre
// File        : Syndrome_ctrl.v
//
//-----------------------------------------------------------------------------
//
// Description : Syndrome module with 4 syndrome cells. Controls the syndrome
//				 calculations
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

//{module {Syndrome_ctrl}}
module Syndrome_ctrl #(parameter width=5)(
	input wire clk,				  //clock signal
	input wire reset,			  //reset signal
	input wire enable,			  //enable signal
	input wire i_start,			  //start syndrome calculations signal
	input wire [width-1:0] s_in,  //symbol input
	output reg S_ready,			  //syndromes ready signal
	output reg NO_ERRORS,		  //no errors signal
	output reg [width-1:0] S1,	  //syndrome value 1
	output reg [width-1:0] S2,	  //syndrome value 2
	output reg [width-1:0] S3,	  //syndrome value 3
	output reg [width-1:0] S4	  //syndrome value 4
	);
	
	//internal constants
	parameter size = 3;
	parameter IDLE = 3'b000;
	parameter STEP1 = 3'b001;
	parameter STEP2 = 3'b011;
	
	//internal variables 
	reg [width-1:0] cnt; //counter max value 31
	reg control;
	
	reg [size-1:0] state;
	reg [size-1:0] next_state;	
	
	//internal wires
	wire [width-1:0] S_1;
	wire [width-1:0] S_2;
	wire [width-1:0] S_3;
	wire [width-1:0] S_4;
	reg [width-1:0] Si;
	
	//module instantiations
	
	//syndrome cell 1
	syndrome_calc_S0 #(.width(width)) SC1(
	.clk	(clk), 
	.reset	(reset),
	.enable	(enable),
	.in		(Si), 		//symbol input
	.ctrl	(control), 	//control signal controls AND gate
	.si		(S_1)		//syndrome value 1 output
	);
	
	//syndrome cell 2
	syndrome_calc_S1 #(.width(width)) SC2(
	.clk	(clk), 
	.reset	(reset),
	.enable	(enable),
	.in		(Si), 	   //symbol input
	.ctrl	(control), //control signal controls AND gate
	.si		(S_2)	   //syndrome value 2 output
	);
	
	//syndrome cell 3
	syndrome_calc_S2 #(.width(width)) SC3(
	.clk	(clk), 
	.reset	(reset),
	.enable	(enable),
	.in		(Si), 	   //symbol input
	.ctrl	(control), //control signal controls AND gate
	.si		(S_3)	   //syndrome value 3 output
	);
	
	//syndrome cell 4
	syndrome_calc_S3 #(.width(width)) SC4(
	.clk	(clk), 
	.reset	(reset),
	.enable	(enable),
	.in		(Si), 	   //symbol input
	.ctrl	(control), //control signal controls AND gate
	.si		(S_4)	   //syndrome value 4 output
	);
	
	
	//--------------FSM-----------------//
	always @(posedge clk or negedge reset)
		begin: FSM_SEQ
			if (reset == 0) begin
				state <= IDLE;
			end else if(enable) begin
				state <= next_state;
			end
		end
	
	//change state section
	always @(state or i_start or cnt)
	begin: FSM_COMBO
		case(state)
			IDLE: begin
				if(i_start == 1) begin
					next_state = STEP1;
				end else begin
					next_state = IDLE;
				end	
			end
			
			STEP1: begin
				if(cnt == 0)begin
					//go back to idle when all symbols shifted in
					//and syndrome values calculated
					next_state = IDLE;	
				end else begin
					next_state = STEP1;
				end
			end
			
			default: next_state = IDLE;		
		endcase
	end
	
	//output section
	always @(posedge clk or negedge reset)
		begin: FSM_OUT
			if(reset == 0)begin
				S_ready <= 0;
				S1 <= 0;
				S2 <= 0;
				S3 <= 0;
				S4 <= 0;
				control <= 0;
				cnt <= 0;
				NO_ERRORS <= 0;
				Si <= 0;
			end else if(enable) begin
				case(state)
					IDLE: begin
						S_ready <= 0;
						control <= 0;
						NO_ERRORS <= 0;
						if(i_start)begin
							//start calculation syndrome values
							cnt <= 30;	//start counter
							Si <= s_in;	//shift in symbol values
						end else begin
							cnt <= 0;
							Si <= 0;
						end
					end
					
					STEP1: begin
						//calcualte all syndrome values
						//takes 31 clock cycles
						if(cnt == 0) begin
							control <= 1;
							cnt <= 0;
							Si <= 0;
							if(S_1 == 0 && S_2 == 0 && S_3 == 0 && S_4 == 0)begin
								//if all syndrome values are zero there are nno errors
								S_ready <= 0;
								S1 <= 0;
								S2 <= 0;
								S3 <= 0;
								S4 <= 0;
								NO_ERRORS <= 1;	//no errors signal set high
							end else begin
								//all syndrome values NOT zero
								//shift out syndrome values
								S1 <= S_1;
								S2 <= S_2;
								S3 <= S_3;
								S4 <= S_4;
								S_ready <= 1;
								NO_ERRORS <= 0;
							end	
						end else begin
							control <= 1;
							cnt <= cnt-1;
							S_ready <= 0;
							NO_ERRORS <= 0;
							S1 <= 0;
							S2 <= 0;
							S3 <= 0;
							S4 <= 0;
							Si <= s_in;	//shift in symbols
						end
					end
					
					default: begin
						S_ready <= 0;
						NO_ERRORS <= 0;
						S1 <= 0;
						S2 <= 0;
						S3 <= 0;
						S4 <= 0;
						control <= 0;
						cnt <= 0;
						Si <= 0;
					end
				endcase
			end
		end
	
endmodule
