//-----------------------------------------------------------------------------
//
// Title       : decoder_top
// Design      : decoder_5bit
// Author      : Sindre
// File        : decoder_top.v
//
//-----------------------------------------------------------------------------
//
// Description : 5-bit RS decoder top module: Configuration 1
//				 Implemented with original Berlekamp-Massey algorithm
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

//{module {decoder_top}}
module decoder_top #(parameter width=5)(
	input clk,					//clock signal
	input reset,				//reset signal
	input enable,				//enable signal 
	input rs_start,				//RS decoding start signal
	input [width-1:0] data_in,	//symbol data input
	output NO_ERRORS,			//no errors detected signal, all syndromes is zero
	output NO_COR,				//high value indicates to many errors to correct
	output rs_done,				//RS decoding done
	output [width-1:0] data_out	//corrected symbol output
	);
	
	//internal regs
	reg sel_1;			//select signal memory selection
	 
	//internal wires
	wire WE_MEM01;
	wire RE_MEM01;
	wire I_START;
	wire S_READY;
	wire L_READY;
	wire LW_READY;
	wire EP_READY;
	wire NOT_VALID;
	wire EV_READY;
	wire RE_INV;
	wire RE1_INV;
	wire RE2_INV;
	
	wire [width-1:0] ADD_WRITE;
	wire [width-1:0] ADD_READ01;
	wire [width-1:0] IN_STAGE_OUT;
	wire [width-1:0] MEM01_OUT;
	wire [width-1:0] S1;
	wire [width-1:0] S2;
	wire [width-1:0] S3;
	wire [width-1:0] S4;
	wire [width-1:0] L0;
	wire [width-1:0] L1;
	wire [width-1:0] L2;
	wire [width-1:0] W0;
	wire [width-1:0] W1;
	wire [width-1:0] ERROR_POS;
	wire [width-1:0] W_TOT;
	wire [width-1:0] L_DER;
	wire [width-1:0] ERROR_VAL;
	
	wire [width-1:0] ADD_READ_INV;
	wire [width-1:0] ADD_READ_I1;
	wire [width-1:0] ADD_READ_I2;
	wire [width-1:0] INV_OUT;
	wire [width-1:0] INV_OUT1;
	wire [width-1:0] INV_OUT2;
	
	
	assign NO_COR = NOT_VALID;
	
	//---------------------------//
	// Module instantiations //
	//-----------------------//
	
	//multiplex control: for reading from inverse memory
	assign RE_INV = sel_1? RE1_INV : RE2_INV;
	assign ADD_READ_INV = sel_1? ADD_READ_I1 : ADD_READ_I2;
	
	//RS decoder input stage
	in_stage #(.width(width)) INPUT_STAGE(
	.clk			(clk),			//clock signal
	.reset			(reset),		//reset signal
	.enable			(enable),		//enable signal
	.rs_start		(rs_start),		//start decoding
	.data_in		(data_in),		//symbol input
	.i_start		(I_START),		//start syndrome calculations signal
	.WE				(WE_MEM01),		//write enable
	.address_write	(ADD_WRITE),	//write address
	.data_out		(IN_STAGE_OUT)	//output bus for in_stage
	);
	
	//Codeword memory storage
	Codeword_Store #(.width(width)) MEM01(
	.clk			(clk),			//clock signal
	.we				(WE_MEM01),		//write enable
	.re				(RE_MEM01),		//read enable
	.address_write	(ADD_WRITE),	//write address
	.address_read	(ADD_READ01),	//read address
	.data_in		(IN_STAGE_OUT),	//memory input
	.data_out		(MEM01_OUT)		//memory output
	);
	
	//Syndrome calculator
	Syndrome_ctrl #(.width(width)) SYNDROME(
	.clk		(clk),			 //clock signal
	.reset		(reset),		 //reset signal
	.enable		(enable),		 //enable signal
	.i_start	(I_START),		 //start syndrome calculations signal
	.s_in		(IN_STAGE_OUT),	 //symbol input
	.S_ready	(S_READY),		 //syndromes ready signal
	.NO_ERRORS	(NO_ERRORS),	 //no errors signal
	.S1			(S1),			 //syndrome value 1
	.S2			(S2),			 //syndrome value 2
	.S3			(S3),			 //syndrome value 3
	.S4			(S4)			 //syndrome value 4
	);
	
	//Berlekamp-Massey algorithm for error locator polynomial
	Original_BM #(.width(width)) O_BM(
	.clk		(clk),			//clock signal
	.reset		(reset),		//reset signal
	.enable		(enable),		//enable signal
	.S_ready	(S_READY),		//syndromes ready signal
	.S1			(S1),			//syndrome value 1
	.S2			(S2),			//syndrome value 2
	.S3			(S3),			//syndrome value 3
	.S4			(S4),			//syndrome value 4
	.INV		(INV_OUT1),		//inverse value input
	.L_ready	(L_READY),		//error locator poly. ready signal
	.RE			(RE1_INV),		//ROM read enable signal
	.VAL		(ADD_READ_I1),	//value to find inverse of
	.L0			(L0),			//locator poly. coefficient 0 output
	.L1			(L1),			//locator poly. coefficient 1 output
	.L2			(L2)			//locator poly. coefficient 2 output
	);
	
	//Error evaluator polynomial calculation
	Omega_calc #(.width(width)) OMEGA(
	.clk		(clk),		//clock signal
	.reset		(reset),	//reset signal
	.enable		(enable),	//enable signal
	.L_ready	(L_READY),	//error locator poly. ready signal
	.S1			(S1),	   	//syndrome value 1
	.S2			(S2),		//syndrome value 2
	.S3			(S3),		//syndrome value 3
	.S4			(S4),		//syndrome value 4
	.L0			(L0),		//locator poly. coefficient 0 input
	.L1			(L1),		//locator poly. coefficient 1 input
	.L2			(L2),		//locator poly. coefficient 2 input
	.LW_ready	(LW_READY),	//error evaluator poly. ready signal
	.W0			(W0),		//evaluator poly. coefficient 0 output
	.W1			(W1)		//evaluator poly. coefficient 1 output
	);
	
	//Chien search
	Chien_search #(.width(width)) CHIEN(
	.clk		(clk),		 //clock signal
	.reset		(reset),	 //reset signal
	.enable		(enable),	 //enable signal
	.LW_ready	(LW_READY),	 //error evaluator poly. ready signal
	.L0			(L0),		 //locator poly. coefficient 0 input
	.L1			(L1),		 //locator poly. coefficient 1 input
	.L2			(L2),		 //locator poly. coefficient 2 input
	.W0			(W0),		 //evaluator poly. coefficient 0 input
	.W1			(W1),		 //evaluator poly. coefficient 1 input
	.EP_ready	(EP_READY),	 //error position ready
	.E_valid	(NOT_VALID), //not valid signal		
	.E_POS		(ERROR_POS), //error position	
	.W_TOT		(W_TOT),	 //Omega(x) value
	.L_DER		(L_DER)		 //error locator polynomial derived
	);
	
	//Forney algorithm
	Forney #(.width(width)) FORNEY(
	.clk		(clk),			//clock signal
	.reset		(reset),		//reset signal
	.enable		(enable),		//enable signal
	.EP_ready	(EP_READY),		//error position ready
	.E_valid	(NOT_VALID),	//valid errors			
	.E_POS		(ERROR_POS),	//error position
	.W_TOT		(W_TOT),	  	//Omega(x) value
	.L_DER		(L_DER),       	//error locator polynomial derived
	.INV		(INV_OUT2),		//inverse value input
	.VAL		(ADD_READ_I2),	//value to find inverse of
	.RE			(RE2_INV),		//read enable		
	.EV_ready	(EV_READY),	    //error value ready signal
	.E_value	(ERROR_VAL)		//error value		    
	);
	
	//ROM storing inverse values of GF(32)
	inverse_GF32 #(.width(width)) INVERSEROM(
	.clk			(clk), 		  	//clock signal
	.re				(RE_INV), 		//read enable
	.address_read	(ADD_READ_INV), //read address
	.data_out		(INV_OUT)		//data output
	);
	
	//MUX for controlling inverse output
	MUX_1to2 #(.width(width)) MUX_INV_OUT(
	.in		(INV_OUT),    //input
	.sel	(sel_1), 	  //select signal
	.out1	(INV_OUT1),	  //output 1
	.out2	(INV_OUT2)	  //output 2
	);
	
	
	//Error correction
	Error_corr #(.width(width)) CORRECTION(
	.clk		(clk),			//clock signal
	.reset		(reset),		//reset signal
	.enable		(enable),		//enable signal
	.EV_ready	(EV_READY),		//error value ready signal
	.E_value	(ERROR_VAL),	//error value
	.S_IN		(MEM01_OUT),	//symbol input from codeword storage
	.RE			(RE_MEM01),		//read enable
	.RS_DONE	(rs_done),		//RS decoding done
	.add_read	(ADD_READ01),	//RAM read address 
	.S_OUT		(data_out)		//corrected symbol value
	);
	
	//controlling select 1 signal based module control signals
	//Controls select siganl for MUX
	always @(posedge clk or negedge reset)
	begin
		if(!reset)begin
			sel_1 <= 0;	
		end else if(enable)begin
			if(S_READY)begin
				sel_1 <= 1;	
			end else if(LW_READY)begin
				sel_1 <= 0;	
			end else begin
				sel_1 <= sel_1;
			end 
		end 
	end
	
endmodule
