//-----------------------------------------------------------------------------
//
// Title       : BM_alg
// Design      : decoder_5bit
// Author      : Sindre
// File        : BM_alg.v
//
//-----------------------------------------------------------------------------
//
// Description : inversionless Berlekamp-Massey algorithm 
//				 calculates error locator polynomial
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

//{module {BM_alg}}
module BM_alg #(parameter width=5)(
	input clk,					//clock signal
	input reset,				//reset signal
	input enable,				//enable signal
	input S_ready,				//syndromes ready signal
	input [width-1:0] S1,		//syndrome value 1
	input [width-1:0] S2,		//syndrome value 2
	input [width-1:0] S3,		//syndrome value 3
	input [width-1:0] S4,		//syndrome value 4
	output reg L_ready,			//error locator poly. ready signal
	output reg [width-1:0] L0,	//locator poly. coefficient 0 output
	output reg [width-1:0] L1,	//locator poly. coefficient 1 output
	output reg [width-1:0] L2	//locator poly. coefficient 2 output
	);
	
	//internal constants
	parameter size = 3;
	//state values are gray coded
	parameter IDLE = 3'b000; 	//0
	parameter STEP1 = 3'b001;	//1 
	parameter STEP2 = 3'b011;	//3
	parameter STEP3 = 3'b010;	//2
	parameter STEP4 = 3'b110;	//6
	parameter STEP5	= 3'b111;	//7
	
	//internal variables //temperary
	reg [width-1:0] L [0:2]; 	// error locator polynomial coefficients
	reg [width-1:0] Lt [0:2];	// temp variable error locator polynomial coefficients
	reg [width-1:0] Li;			// error locator single coeff
	reg [width-1:0] T [0:2]; 	// auxiliary polynomial (tau)
	reg [width-1:0] D;			// discrepancy (uppercase Delta)
	reg [width-1:0] Dt;			// temp next dicrepancy (uppercase Delta D^(i+1))
	reg [width-1:0] d_prev;		// previous non-zero discrepancy (lowecase delta)
	reg [width-1:0]	K;			// auxiliary degree variable
	reg [2:0] N;				//counter, max value 4 (Iteration counter)
	reg [1:0] i;				//counter, max value 2 (ith step counter)
	
	reg [width-1:0] S [1:4];	// internal storage for syndrome values
	
	reg [size-1:0] state;
	reg [size-1:0] next_state;	
	
	//input variables to modules
	reg [width-1:0] M01;		//M<value> indicates an input to a multiplier
	reg [width-1:0] M02;
	reg [width-1:0] M03;
	reg [width-1:0] M04;
	reg [width-1:0] M05;
	reg [width-1:0] M06;
	reg [width-1:0] A01;	   //A<value> indicates an input to a adder
	//internal wires
	wire [width-1:0] NET01;	   //NET<value> indicates an internal wire
	wire [width-1:0] NET02;
	wire [width-1:0] NET03;
	wire [width-1:0] NET04;
	wire [width-1:0] SUM01;	   //SUM<value> indicates an output from an multiplier or adder
	wire [width-1:0] SUM02;
	
	
	//module instantiations
	//modules used for performing finite field addition and multiplication
	Full_mult #(.width(width)) FFM01(
	.A	(M01),
	.B	(M02),
	.Y	(NET01)
	);
	
	Full_mult #(.width(width)) FFM02(
	.A	(M03),
	.B	(M04),
	.Y	(NET02)
	);
	
	Full_mult #(.width(width)) FFM03(
	.A	(M05),
	.B	(M06),
	.Y	(NET03)
	);
	
	Adder #(.width(width)) FFA01(
	.in1	(NET01),
	.in2	(A01),
	.sum	(SUM01)
	);
   
	Adder #(.width(width)) FFA02(
	.in1	(NET02),
	.in2	(NET03),
	.sum	(SUM02)
	);
	
	//FSM
	always @(posedge clk or negedge reset)
		begin: FSM_SEQ
			if (reset == 0) begin
				state <= IDLE;
			end else if(enable) begin
				state <= next_state;
			end
		end
	
	//change state section
	always @(state or S_ready or N or i)
	begin: FSM_COMBO
		case(state)
			IDLE: begin
				if(S_ready)begin
					//start calculating when syndromes ready
					next_state = STEP1;
				end else begin
					next_state = IDLE;
				end
			end
			
			STEP1: begin
				next_state = STEP3;
			end
			
			STEP2: begin
				next_state = STEP3;
			end
			
			STEP3: begin
				if(N == 0)begin
					if(i == 2)begin
						next_state = STEP4;
					end else begin
						next_state = STEP2;
					end
				end else begin
					if(i == 3) begin
						next_state = STEP4;
					end else begin
						next_state = STEP2;
					end
				end		
			end
			
			STEP4: begin
				if(N<=2) begin
					//continue if not done
					next_state = STEP1;
				end else begin
					//if N=3 the calculations are finished
					next_state = STEP5;
				end				
			end
	
			STEP5: begin
				//when calculaton done, go back to IDLE
				next_state = IDLE;
			end
			
			default: next_state = IDLE;
		endcase	
	end
	
	//output section
	always @(posedge clk or negedge reset)
		begin: FSM_OUT
			if(reset == 0)begin
				//reset all internal regs
				L_ready <= 0;
				L[0] <= 0;
				L[1] <= 0;
				L[2] <= 0;
				Lt[0] <= 0;
				Lt[1] <= 0;
				Lt[2] <= 0;
				S[1] <= 0; S[3] <= 0;
				S[2] <= 0; S[4] <= 0;
				T[0] <= 0; T[2] <= 0;
				T[1] <= 0;
				Li <= 0;
				D <= 0;
				Dt <= 0;
				d_prev <= 0;
				K <= 0;
				N <= 0;
				i <= 0;
				L0 <= 0;
				L1 <= 0;
				L2 <= 0;
				A01 <= 0;
				M01 <= 0;
				M02 <= 0;
				M03 <= 0;
				M04 <= 0;
				M05 <= 0;
				M06 <= 0;
			end else if(enable) begin
				case(state)
					IDLE: begin
						//set initial values
						L_ready <= 0;
						L[0] <= 1;
						L[1] <= 0;
						L[2] <= 0;
						Lt[0] <= 1;
						Lt[1] <= 0;
						Lt[2] <= 0;
						T[0] <= 1; 
						T[1] <= 0;
						T[2] <= 0;
						d_prev <= 1;
						K <= 0;
						N <= 0;
						i <= 0;
						if(S_ready)begin
							//syndromes ready start calulation of
							//error locator poly.
							S[1] <= S1;
							S[2] <= S2;
							S[3] <= S3;
							S[4] <= S4;
							D <= S1;
						end else begin
							S[1] <= 0; S[3] <= 0;
							S[2] <= 0; S[4] <= 0;
							D <= 0; 
						end
					end
					
					STEP1: begin
						//first iteration
						M01 <= 0;
						M02 <= 0;
						A01 <= 0;
						M03 <= D;
						M04 <= 0;
						M05 <= d_prev;
						M06 <= L[0];
					end
					
					STEP2: begin
						if(N == 2*2-1) begin
							// No calculate of discrepancy in last cycle
							M01 <= 0;
							M02 <= 0;	
							A01 <= 0;
									
							//Calculate error locator poly. coeff 
							M03 <= D;
							M04 <= T[i-1];
							M05 <= d_prev;
							M06 <= L[i];
							
						end else begin
							if(i == 3)begin
								//Calculate discrepancy
								M01 <= S[N-i+3];
								M02 <= Lt[i-1];	
								A01 <= Dt;
							end else begin
								//Calculate discrepancy
								M01 <= S[N-i+3];
								M02 <= Lt[i-1];	
								A01 <= Dt;		
								//Calculate error locator coeff 
								M03 <= D;
								M04 <= T[i-1];
								M05 <= d_prev;
								M06 <= L[i];
							end	
						end
					end
					
					STEP3: begin
						//read calculated values
						Dt <= SUM01;
						Lt[i] <= SUM02;
						i <= i+1;
					end
					
					STEP4: begin
						//REF see line 4 to 8 in iBM algorithm (algorithm 2 in thesis)
						if((D == 0) || (2*K >= N+1)) begin
							//update reg values
							K <= K;
							d_prev <= d_prev;
							T[0] <= 0;
							T[1] <= T[0];
							T[2] <= T[1];
							L[0] <= Lt[0];
							L[1] <= Lt[1];
							L[2] <= Lt[2];
						end else begin
							//update reg values
							K <= N+1-K;
							d_prev <= D;
							T[0] <= L[0];
							T[1] <= L[1];
							T[2] <= L[2];
							L[0] <= Lt[0];
							L[1] <= Lt[1];
							L[2] <= Lt[2];
						end
						D <= Dt;	//discrepancy D^(i+1)
						N <= N+1;	//increase iteration counter
						i <= 0;		//reset the step counter
					end
					
					STEP5: begin
						//shift out error locator coeff
						L_ready <= 1;
						L0 <= Lt[0];
						L1 <= Lt[1];
						L2 <= Lt[2];
					end
					
					default: begin
						L_ready <= 0;
						L[0] <= 0;
						L[1] <= 0;
						L[2] <= 0;
						Lt[0] <= 0;
						Lt[1] <= 0;
						Lt[2] <= 0;
						S[1] <= 0; S[3] <= 0;
						S[2] <= 0; S[4] <= 0;
						T[0] <= 0; T[2] <= 0;
						T[1] <= 0;
						Li <= 0;
						D <= 0;
						Dt <= 0;
						d_prev <= 0;
						K <= 0;
						N <= 0;
						i <= 0;
						L0 <= 0;
						L1 <= 0;
						L2 <= 0;
						A01 <= 0;
						M01 <= 0;
						M02 <= 0;
						M03 <= 0;
						M04 <= 0;
						M05 <= 0;
						M06 <= 0;
					end
				endcase
			end
		end
	
endmodule
