//-----------------------------------------------------------------------------
//
// Title       : Forney
// Design      : decoder_5bit
// Author      : Sindre
// File        : Forney.v
//
//-----------------------------------------------------------------------------
//
// Description : Forney's method for finding error values
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

//{module {Forney}}
module Forney #(parameter width=5)(
	input clk,				    //clock signal
	input reset,				//reset signal
	input enable,				//enable signal
	input EP_ready,				//error position ready
	input E_valid,				//error valid, high value indicates too may errors to correct
	input [width-1:0] E_POS,	//error position
	input [width-1:0] W_TOT,	//omega(x) values
	input [width-1:0] L_DER,    //error locator polynomial derived
	input [width-1:0] INV,		//inverse value input
	output [width-1:0] VAL,		//value to find inverse of
	output reg RE,				//read enable to INVERSE ROM
	output reg EV_ready,	    //error value ready signal
	output [width-1:0] E_value	//error value
	);
	
	//internal constants
	parameter size = 3;
	parameter IDLE = 3'b000;
	parameter STEP1 = 3'b001;
	
	//internal variables
	reg [4:0] cnt;				//counter: max 31 - number of symbols
	reg [size-1:0] state;
	reg [size-1:0] next_state;
	
	//internal wires and reg
	wire [width-1:0] M_OUT;		//output of	full multiplier
	wire [width-1:0] R_OUT01;
	wire [width-1:0] R_OUT03;
	wire [width-1:0] R_OUT04;
	wire [width-1:0] R_OUT05;
	
	//module instantiations
		
	//AND error position and value so only 
	//valid error values are send to error correction module
	assign E_value = M_OUT & R_OUT05; 
	
	//registers used to delay values why inverse is found
	D_reg #(.width(width)) REG01(
	.clk	(clk), 
	.reset	(reset), 
	.enable	(enable), 
	.d_in	(W_TOT), 
	.d_out	(R_OUT01)
	);
	
	D_reg #(.width(width)) REG02(
	.clk	(clk), 
	.reset	(reset), 
	.enable	(enable), 
	.d_in	(L_DER), 
	.d_out	(VAL)
	);
	
	D_reg #(.width(width)) REG03(
	.clk	(clk), 
	.reset	(reset), 
	.enable	(enable), 
	.d_in	(E_POS), 
	.d_out	(R_OUT03)
	);
	
	D_reg #(.width(width)) REG04(
	.clk	(clk), 
	.reset	(reset), 
	.enable	(enable), 
	.d_in	(R_OUT01), 
	.d_out	(R_OUT04)
	);
	
	D_reg #(.width(width)) REG05(
	.clk	(clk), 
	.reset	(reset), 
	.enable	(enable), 
	.d_in	(R_OUT03), 
	.d_out	(R_OUT05)
	);
	
	//multiply inverse value of locator poly derived and omega(x)
	Full_mult #(.width(width)) FFM01(
	.A	(R_OUT04),
	.B	(INV),		
	.Y	(M_OUT)
	);
	
	//FSM
	always @ (posedge clk or negedge reset)
		begin: FSM_SEQ
			if (reset == 0)begin
				state <= IDLE;
			end else if(enable) begin
				state <= next_state;
			end
		end
	//state change section
	always @ (state or EP_ready or E_valid or cnt)
		begin: FSM_COMBO
			case(state)
				IDLE: begin
					if(EP_ready)begin
						next_state = STEP1;
					end else begin
						next_state = IDLE;
					end 
				end
						
				STEP1: begin
					if(E_valid == 1)begin
						next_state = IDLE;	
					end else begin
						if(cnt == 0)begin
							//go back to IDLE when done
							next_state = IDLE;
						end else begin
							next_state = STEP1;
						end
					end
				end
				
				default: next_state = IDLE; 
			endcase
		end	
	
	//output section
	always @ (posedge clk or negedge reset)
		begin: FSM_OUT
			if(reset == 0)begin
				cnt <= 0;
				RE <= 0;				
				EV_ready <= 0;
			end else if(enable) begin
				case(state)
					IDLE: begin
						if(EP_ready == 1)begin
							//error position ready, start calculating error values
							RE <= 1;			
							cnt <= 31;
							EV_ready <= 1; //error value ready
						end else begin
							EV_ready <= 0;
							RE <= 0;
							cnt <= 0;
						end
					end
						
					STEP1: begin
						EV_ready <= 0; //"error value ready signal" only high one cc
						if(E_valid == 1)begin
							//as long as valid continue
							RE <= 0;
						end else begin
							RE <= 1;
							//calculate for next position
							cnt <= cnt-1;
						end
					end
					
					default: begin
						RE <= 0;				
						EV_ready <= 0;
						cnt <= 0;
					end
				endcase
			end
		end

endmodule
