//-----------------------------------------------------------------------------
//
// Title       : Syndrome_ctrl
// Design      : decoder_5bit
// Author      : Sindre
// File        : Syndrome_ctrl.v
//
//-----------------------------------------------------------------------------
//
// Description : Computes the first two syndromes (S1 and S2). If both 
//				 syndromes is zero, the code word is assumed error free.
//				 If the two syndromes is not equal zero, the remaining 
//				 syndromes (S3 and S4) are calculated
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

//{module {Syndrome_ctrl}}
module Syndrome_ctrl #(parameter width=5)(
	input wire clk,						//clock signal
	input wire reset,					//reset signal
	input wire enable,					//enable signal
	input wire i_start,					//Syndrome computation start signal
	input wire [width-1:0] s_in,		//symbol input
	input wire [width-1:0] mem_in,		//input from codeword memory
	output reg RE,						//read enable for reading from codeword memory
	output reg S_ready,					//syndrome values ready
	output reg NO_ERRORS,				//no errors flag
	output reg [width-1:0] address_read,//memory address read
	output reg [width-1:0] S1,			//syndrome value 1
	output reg [width-1:0] S2,			//syndrome value 2
	output reg [width-1:0] S3,			//syndrome value 3
	output reg [width-1:0] S4			//syndrome value 4
	);
	
	
	//internal constants
	parameter size = 3;
	parameter IDLE = 3'b000;
	parameter STEP1 = 3'b001;
	parameter STEP2 = 3'b011;
	
	//internal variables 
	reg [width-1:0] cnt; //counter max value 31
	reg control;
	
	reg [size-1:0] state;
	reg [size-1:0] next_state;	
	
	//internal wires
	wire [width-1:0] S_1;
	wire [width-1:0] S_2;
	wire [width-1:0] S_3;
	wire [width-1:0] S_4;
	wire gclk1;
	wire gclk2;
	
	reg gclk;
	reg qint;
	
	reg [width-1:0] Si;
	reg en1;			//enable signal for SC1 and SC2
	reg en2;			//enable signal for SC3 and SC4
	
	//module instantiations
	clk_gate CG1(
	.clk	(clk), 
	.ctrl	(en1), 
	.gclk	(gclk1)
	);
	
	clk_gate CG2(
	.clk	(clk), 
	.ctrl	(en2), 
	.gclk	(gclk2)
	);
	
	//syndrome cells
	syndrome_calc_S0 #(.width(width)) SC1(
	.clk	(gclk1), 
	.reset	(reset),
	.enable	(en1),
	.in		(Si), 	  
	.ctrl	(control), 
	.si		(S_1)
	);
	syndrome_calc_S1 #(.width(width)) SC2(
	.clk	(gclk1), 
	.reset	(reset),
	.enable	(en1),
	.in		(Si), 
	.ctrl	(control), 
	.si		(S_2)
	);
	
	syndrome_calc_S2 #(.width(width)) SC3(
	.clk	(gclk2), 
	.reset	(reset),
	.enable	(en2),
	.in		(mem_in), 
	.ctrl	(control), 
	.si		(S_3)
	);
	
	syndrome_calc_S3 #(.width(width)) SC4(
	.clk	(gclk2), 
	.reset	(reset),
	.enable	(en2),
	.in		(mem_in), 
	.ctrl	(control), 
	.si		(S_4)
	);
	
	always@(clk or enable)
	begin
		if(clk == 0)begin
			qint <= (!enable);
		end
		
		gclk <= (!qint) && clk;
	end
	
	//FSM
	always @(posedge gclk or negedge reset)
		begin: FSM_SEQ
			if (reset == 0) begin
				state <= IDLE;
			end else if(enable) begin
				state <= next_state;
			end
		end
	
	//change state section
	always @(state or i_start or cnt or S_1 or S_2)
	begin: FSM_COMBO
		case(state)
			IDLE: begin
				if(i_start == 1) begin
					next_state = STEP1;
				end else begin
					next_state = IDLE;
				end	
			end
			
			STEP1: begin
				if(cnt == 0)begin
					if(S_1 == 0 && S_2 == 0)begin
						//if both values zero, go back to IDLE
						next_state = IDLE;
					end else begin
						//if not zero, continue to next step
						next_state = STEP2;
					end
				end else begin
					//as long as not done, stay in state
					next_state = STEP1;
				end
			end
	
			STEP2: begin
				if(cnt == 0)begin
					//go back to idle when counter is zero
					next_state = IDLE;
				end else begin
					next_state = STEP2;
				end
			end
			
			default: next_state = IDLE;		
		endcase
	end
	
	//output section
	always @(posedge gclk or negedge reset)
		begin: FSM_OUT
			if(reset == 0)begin
				S_ready <= 0;
				S1 <= 0;
				S2 <= 0;
				S3 <= 0;
				S4 <= 0;
				Si <= 0;
				en1 <= 0;
				en2 <= 0;
				control <= 0;
				cnt <= 0;
				NO_ERRORS <= 0;
				RE <= 0;
				address_read <= 0;
			end else if(enable) begin
				case(state)
					IDLE: begin
						S_ready <= 0;
						control <= 0;
						NO_ERRORS <= 0;
						RE <= 0;
						address_read <= 0;
						if(i_start)begin
							//Start to calculate the first two syndrome values
							cnt <= 30;
							Si <= s_in;
							en1 <= 1;
							en2 <= 0;
						end else begin
							cnt <= 0;
							Si <= 0;
							en1 <= 0;
							en2 <= 0;
						end
					end
					
					STEP1: begin
						if(cnt == 0) begin
							control <= 1;
							Si <= 0;
							if(S_1 == 0 && S_2 == 0)begin
								//check if the two first syndromes are zero
								//if zero, no errors
								cnt <= 0;
								S_ready <= 0;
								S1 <= 0;
								S2 <= 0;
								S3 <= 0;
								S4 <= 0;
								NO_ERRORS <= 1;
								RE <= 0;
								address_read <= 0;
								en1 <= 0;
								en2 <= 0;
							end else begin
								//if not zero calculate rest of syndrome values
								//start reading symbols from memory
								cnt <= 31;
								S1 <= S_1;
								S2 <= S_2;
								S3 <= 0;
								S4 <= 0;
								S_ready <= 0;
								NO_ERRORS <= 0;
								RE <= 1;
								address_read <= 30;
								en1 <= 0;
								en2 <= 1;
							end	
						end else begin
							RE <= 0;
							control <= 1;
							cnt <= cnt-1;
							S_ready <= 0;
							NO_ERRORS <= 0;
							S1 <= 0;
							S2 <= 0;
							S3 <= 0;
							S4 <= 0;
							Si <= s_in;	//input symbols into syndrome cells
						end
					end
					
					STEP2: begin
						//calculate rest of syndrome values
						if(cnt == 0)begin	
							S_ready <= 1;	//syndrome calulations done
							cnt <= 0;
							RE <= 0;
							address_read <= 0;
							S3 <= S_3;	  	//output the calculated syndrome value
							S4 <= S_4;		//output the calculated syndrome value
						end else begin
							if(cnt == 31)begin
								control <= 0;
							end else begin
								control <= 1;
							end
							if(cnt == 1)begin
								RE <= 0;
								address_read <= 0;
							end else begin
								RE <= 1;
								address_read <= cnt-2;	//read symbols from memory
							end
							S_ready <= 0;
							cnt <= cnt-1;  //decrease counter
							S3 <= 0;
							S4 <= 0;
						end
					end
					
					default: begin
						RE <= 0;
						S_ready <= 0;
						NO_ERRORS <= 0;
						S1 <= 0;
						S2 <= 0;
						S3 <= 0;
						S4 <= 0;
						Si <= 0;
						en1 <= 0;
						en2 <= 0;
						control <= 0;
						cnt <= 0;
						Si <= 0;
					end
				endcase
			end
		end
	
endmodule
