//-----------------------------------------------------------------------------
//
// Title       : in_stage
// Design      : decoder_5bit
// Author      : Sindre
// File        : in_stage.v
//
//-----------------------------------------------------------------------------
//
// Description : RS decoder input stage, shift in symbols from codeword
//				 5bit at the time
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

//{module {in_stage}}
module in_stage #(parameter width=5)(
	input clk,						//clock signal
	input reset,					//reset signal
	input enable,					//enable signal
	input rs_start,					//start deconding signal
	input [width-1:0] data_in,		//data symbol input
	output reg i_start,				//start syndrome calculations signal
	output reg WE,					//write enable (memory)
	output reg [width-1:0] address_write, //write address
	output reg [width-1:0] data_out		  //data output
	);
	
	//internal constants
	parameter size = 3;
	parameter IDLE = 3'b000;
	parameter STEP1 = 3'b001;

	//internal variables
	reg [4:0] cnt;				//counter: max 31 - number of symbols
	reg [size-1:0] state;
	reg [size-1:0] next_state;
	
	reg gclk;
	reg qint;
	
	//clock gating
	always@(clk or enable)
	begin
		if(clk == 0)begin
			qint <= (!enable);
		end
		gclk <= (!qint) && clk;
	end
	
	//FSM
	always @(posedge gclk or negedge reset)
	begin: FSM_SEQ
		if(reset == 0)begin	  
			state <= IDLE;
		end else if(enable)begin
			state <= next_state;	
		end
	end
	
	//change state section
	always @(state or rs_start or cnt)
	begin: FSM_COMBO
		case(state)
			IDLE: begin
				if(rs_start)begin
					//start to shift in symbols
					next_state = STEP1;	
				end else begin
					next_state = IDLE;
				end
			end
			
			STEP1: begin
				if(cnt == 0)begin
					//when all symbols shifted in, go back to IDLE
					next_state = IDLE;
				end else begin
					next_state = STEP1;
				end
			end
			
			default: next_state = IDLE;
		endcase
	end
	
	//output section
	always @(posedge gclk or negedge reset)
	begin: FSM_OUT
		if(reset == 0)begin
			//reset register and outputs
			i_start <= 0;
			WE <= 0;
			address_write <= 0;
			data_out <= 0;
			cnt <= 0;
		end else if(enable) begin
			case(state)
				IDLE: begin
					if(rs_start)begin
						//start in shifting
						i_start <= 1;	//start syndrome calculations
						WE <= 1;
						address_write <= 30;  //write to first address
						data_out <= data_in; //output symbol
						cnt <= 29;
					end else begin
						i_start <= 0;
						WE <= 0;
						address_write <= 0;
						data_out <= 0;
						cnt <= 0;
					end
				end
				
				STEP1: begin
					//shift in rest of symbols
					//and write to memory and send to syndrome module
					WE <= 1;
					address_write <= cnt;
					data_out <= data_in;
					i_start <= 0;
					if(cnt == 0)begin
						cnt <= 0;
					end else begin
						cnt <= cnt-1;
					end
				end
				
				default: begin
		  			i_start <= 0;
					WE <= 0;
					address_write <= 0;
					data_out <= 0;
					cnt <= 0;
				end
			endcase
		end
	end
	
endmodule
