//-----------------------------------------------------------------------------
//
// Title       : in_stage
// Design      : decoder_5bit
// Author      : Sindre
// File        : in_stage.v
//
//-----------------------------------------------------------------------------
//
// Description : RS decoder input stage, shift in symbols from codeword
//				 5bit at the time
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

//{module {in_stage}}
module in_stage #(parameter width=5)(
	input clk,							   //clock signal
	input reset,						   //reset signal
	input enable,						   //enable signal
	input rs_start,						   //start decoding
	input [width-1:0] data_in,			   //symbol input
	output reg i_start,					   //start syndrome calculations signal
	output reg WE,						   //write enable
	output reg [width-1:0] address_write,  //write address
	output reg [width-1:0] data_out		   //output bus for in_stage
	);
	
	//internal constants
	parameter size = 3;
	parameter IDLE = 3'b000;
	parameter STEP1 = 3'b001;

	//internal variables
	reg [4:0] cnt;				//counter: max 31 - number of symbols
	reg [size-1:0] state;
	reg [size-1:0] next_state;
	
	reg gclk;
	reg qint;
	
	//clock gating
	always@(clk or enable)
	begin
		if(clk == 0)begin
			qint <= (!enable);
		end
		
		gclk <= (!qint) && clk;
	end
	
	//FSM
	always @(posedge gclk or negedge reset)
	begin: FSM_SEQ
		if(reset == 0)begin	  
			state <= IDLE;
		end else if(enable)begin
			state <= next_state;	
		end
	end
	
	//change state section
	always @(state or rs_start or cnt)
	begin: FSM_COMBO
		case(state)
			IDLE: begin
				if(rs_start)begin
					next_state = STEP1;	
				end else begin
					next_state = IDLE;
				end
			end
			
			STEP1: begin
				if(cnt == 0)begin
					//go back to IDLE state when all symbols are shifted in
					next_state = IDLE;
				end else begin
					next_state = STEP1;
				end
			end
			
			default: next_state = IDLE;
		endcase
	end
	
	//output section
	always @(posedge gclk or negedge reset)
	begin: FSM_OUT
		if(reset == 0)begin
			i_start <= 0;
			WE <= 0;
			address_write <= 0;
			data_out <= 0;
			cnt <= 0;
		end else if(enable) begin
			case(state)
				IDLE: begin
					if(rs_start)begin
						//start in shifting of symbols to memory
						i_start <= 0;
						WE <= 1;
						address_write <= 30; //write symbols to memory
						data_out <= data_in; //shift symbols to syndrome and memody modules
						cnt <= 29;
					end else begin
						i_start <= 0;
						WE <= 0;
						address_write <= 0;
						data_out <= 0;
						cnt <= 0;
					end
				end
				
				STEP1: begin
					//shift in rest of symbols and store in memory
					WE <= 1;
					address_write <= cnt;
					data_out <= data_in;
					if(cnt == 17)begin
						//when half of symbols shifted in, start syndrome calculations
						cnt <= cnt-1;
						i_start <= 1;
					end else if(cnt == 0)begin
						i_start <= 0;
						cnt <= 0;
					end else begin
						cnt <= cnt-1;
						i_start <= 0;
					end
				end
				
				default: begin
		  			i_start <= 0;
					WE <= 0;
					address_write <= 0;
					data_out <= 0;
					cnt <= 0;
				end
			endcase
		end
	end
	
endmodule
