//-----------------------------------------------------------------------------
//
// Title       : Modified_Chien_search
// Design      : decoder_5bit
// Author      : Sindre
// File        : Modified_Chien_search.v
//
//-----------------------------------------------------------------------------
//
// Description : Modified Chien search module. Computes all lamdba(x) values
//				 first, and stores only the positions where there is an error.
//				 No pipelining when using this module. Reduces the number of
//				 computations in the Forney module since only valid values are
//				 passed through
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

//{module {Modified_Chien_search}}
module Modified_Chien_search #(parameter width=5) (
	input clk,				//clock signal
	input reset,			//reset signal
	input enable,			//enable signal
	input LW_ready,			//error evaluator poly. ready signal
	input [width-1:0] L0,	//locator poly. coefficient 0 input
	input [width-1:0] L1,	//locator poly. coefficient 1 input
	input [width-1:0] L2,	//locator poly. coefficient 2 input
	input [width-1:0] W0,	//evaluator poly. coefficient 0 input
	input [width-1:0] W1,	//evaluator poly. coefficient 1 input
	output reg NO_COR,				//no correction flag
	output reg EP_ready,			//error position ready
	output [1:0] E_valid,			//number of valid errors
	output reg [width-1:0] POS,		//real error position output
	output reg [width-1:0] E_POS,	//error position
	output reg [width-1:0] W_TOT,	//outputs the calculated omega(x) value at each clock cycle
	output reg [width-1:0] L_DER	//lamda derivative (lamda odd), to use in Forney alg
	);
	
	//internal constants
	parameter A0 = 5'b00001;	//decimal value 1
	parameter A1 = 5'b00010;	//decimal value 2
	parameter A2 = 5'b00100;	//decimal value 4
	
	parameter size = 3;
	parameter IDLE = 3'b000;
	parameter STEP1 = 3'b001;
	parameter STEP2 = 3'b011;
	parameter STEP3 = 3'b010;
	
	//internal variables
	reg [width-1:0] E [0:1];	//internal storage for error positions
	reg [width-1:0] Ld [0:1];	//Lamda derivative values
	reg [width-1:0] wt [0:1];	//omega(x) values
	reg [width-1:0] cnt;		//counter: max 31 - number of symbols
	reg [size-1:0] state;
	reg [size-1:0] next_state;
	reg [1:0] e_cnt;			//counter used for shifting out error values
	reg [1:0] errors;			//counter to track number of error positions, max 2
	reg ctrl;					//Control signal for selecting input on multiplexers
	reg tm;						//too many errors flag, if set high indicates that there are more the two errors
	reg gclk;					//gated clock signal
	reg qint;
	
	//internal wires
	wire [width-1:0] OUT_CELL00;
	wire [width-1:0] OUT_CELL01;
	wire [width-1:0] OUT_CELL02;
	wire [width-1:0] OUT_CELL03;
	wire [width-1:0] OUT_CELL04;
	wire [width-1:0] OUT_FFA01;
	wire [width-1:0] OUT_FFA02;
	wire [width-1:0] OUT_FFA03;
	
	//assigns
	assign E_valid = errors;
	
	//-------------------------//
	//  Module instantiations  //
	//-------------------------//
	
	//calculation of lamda(x) values
	Chien_cell #(.width(width)) CC00(
	.clk	(gclk),		 //clock signal
	.reset	(reset),	 //reset signal
	.enable	(enable),	 //enable signal
	.sel	(ctrl),		 //select signal
	.L		(L0),		 //error locator coefficient 0 input
	.a		(A0),		 //generator root alpha^0 input
	.out	(OUT_CELL00) //output
	);
	
	Chien_cell #(.width(width)) CC01(
	.clk	(gclk),		 //clock signal
	.reset	(reset),	 //reset signal
	.enable	(enable),	 //enable signal
	.sel	(ctrl),		 //select signal
	.L		(L1),		 //error locator coefficient 1 input
	.a		(A1),		 //generator root alpha^1 input
	.out	(OUT_CELL01) //output
	);
	
	Chien_cell #(.width(width)) CC02(
	.clk	(gclk),		//clock signal
	.reset	(reset),	//reset signal
	.enable	(enable),	//enable signal
	.sel	(ctrl),		//select signal
	.L		(L2),		//error locator coefficient 2 input
	.a		(A2),		//generator root alpha^2 input
	.out	(OUT_CELL02)//output
	);
	
	//add the output from cells
	Adder #(.width(width)) FFA01(
	.in1	(OUT_CELL00),
	.in2	(OUT_CELL01),
	.sum	(OUT_FFA01)
	);
	
	Adder #(.width(width)) FFA02(
	.in1	(OUT_FFA01),
	.in2	(OUT_CELL02),
	.sum	(OUT_FFA02)
	);
	
	//calculation of omega(x) values
	Chien_cell_W #(.width(width)) CCW00(
	.clk	(gclk),		//clock signal
	.reset	(reset),	//reset signal
	.enable	(enable),	//enable signal
	.sel	(ctrl),		//select signal
	.W		(W0),		//error evaluator coefficient 0 input
	.a		(A0),		//generator root alpha^0 input
	.out	(OUT_CELL03)//output
	);
	
	Chien_cell_W #(.width(width)) CCW01(
	.clk	(gclk),		 //clock signal
	.reset	(reset),	 //resey signal
	.enable	(enable),	 //enable signal
	.sel	(ctrl),		 //select signal
	.W		(W1),		 //error evaluator coefficient 0 input
	.a		(A1),		 //generator root alpha^1 input
	.out	(OUT_CELL04) //output
	);
	
	Adder #(.width(width)) FFA03(
	.in1	(OUT_CELL03),
	.in2	(OUT_CELL04),
	.sum	(OUT_FFA03)
	);
	
	//clock gating
	always@(clk or enable)
	begin
		if(clk == 0)begin
			qint <= (!enable);
		end
		
		gclk <= (!qint) && clk;
	end
	
	//FSM
	always @ (posedge gclk or negedge reset)
		begin: FSM_SEQ
			if (reset == 0)begin
				state <= IDLE;
			end else if(enable) begin
				state <= next_state;
			end
		end
	//state selection section
	always @ (state or LW_ready or cnt or tm or e_cnt or errors)
		begin: FSM_COMBO
			case(state)
				IDLE: begin
					if(LW_ready == 1) begin
						//start chien search
						next_state = STEP1;
					end else begin
						next_state = IDLE;
					end	
				end
						
				STEP1: begin
					//check all positions
					if(tm == 1)begin
						next_state = IDLE;
					end else begin
						if(cnt == 0) begin
							//when all 31 symbols checked, continue to next step
							next_state = STEP2;
						end else begin
							next_state = STEP1;
						end
					end
				end
	
				STEP2: begin
					if(errors == 0)begin
						next_state = IDLE;
					end else begin
						next_state = STEP3;
					end
				end
				
				//shift out errors
				STEP3: begin
					if(e_cnt < errors)begin
						next_state = STEP3;
					end else begin
						//go back to idle when all error pos. shifted out
						next_state = IDLE;	
					end	
				end
							
				default: next_state = IDLE; 
			endcase
		end	
	
	//output section
	always @ (posedge gclk or negedge reset)
		begin: FSM_OUT
			if(reset == 0)begin
				//reset all registers
				NO_COR <= 0;
				EP_ready <= 0;
				POS <= 0;
				E_POS <= 0;
				W_TOT <= 0;
				L_DER <= 0;
				E[0] <= 0;
				E[1] <= 0;
				Ld[0] <= 0;		//Lamda derivative is set to zero
				Ld[1] <= 0;		
				wt[0] <= 0;
				wt[1] <= 0;
				tm <= 0;
				errors <= 0;	//Number of errors set to zero
				cnt <= 0;		//Reset counter to zero
				e_cnt <= 0;		//Reset counter
				ctrl <= 0;		//Lambda and omega coeff is selected on the MUX in Chien cell
			end else if(enable) begin
				case(state)
					IDLE: begin
						NO_COR <= 0;
						EP_ready <= 0;
						POS <= 0;
						E_POS <= 0;
						W_TOT <= 0;
						L_DER <= 0;
						E[0] <= 0;
						E[1] <= 0;
						Ld[0] <= 0;
						Ld[1] <= 0;
						wt[0] <= 0;
						wt[1] <= 0;
						tm <= 0;
						errors <= 0;
						e_cnt <= 0;
						if(LW_ready) begin
							//start chien search
							cnt <= 30;
							ctrl <= 1;
						end else begin
							cnt <= 0;
							ctrl <= 0;
						end	
					end
						
					STEP1: begin
						//check all symbol positions
						if(cnt == 0)begin
							ctrl <= 0;
							cnt <= 0;
						end else begin
							ctrl <= 1;
							cnt <= cnt-1;
						end 
						if(OUT_FFA02 == 0) begin
							//if value is zero,, there is an error on that position
							if(errors == 2)begin
								tm <= 1;
								errors <= 0;
								E[0] <= 0;
								E[1] <= 0;
								Ld[0] <= 0;
								Ld[1] <= 0;
								wt[0] <= 0;
								wt[1] <= 0;
							end else begin
								tm <= 0;
								E[errors] <= cnt;			//store error position
								Ld[errors] <= OUT_CELL01;	//store lambda derivative value
								wt[errors] <= OUT_FFA03;	//store omega value
								errors <= errors+1;			//increment error counter
							end
						end else begin
							tm <= 0;
							E[errors] <= 0;
							Ld[errors] <= 0;
							wt[errors] <= 0;
							errors <= errors;
						end
					end
					
					STEP2: begin
						if(errors == 0)begin
							NO_COR <= 1;
							EP_ready <= 0;
						end else begin
							NO_COR <= 0;
							EP_ready <= 1;
						end 	
					end
	
					STEP3: begin
						if(e_cnt < errors)begin
							//shift out all found error positions
							EP_ready <= 1;
							POS <= E[e_cnt];
							E_POS <= 5'b11111;
							W_TOT <= wt[e_cnt];
							L_DER <= Ld[e_cnt];
							e_cnt <= e_cnt+1;
						end else begin
							EP_ready <= 0;
							POS <= 0;
							E_POS <= 0;
							W_TOT <= 0;
							L_DER <= 0;
							e_cnt <= 0;
						end	
					end
					
					default: begin
						NO_COR <= 0;
						EP_ready <= 0;
						POS <= 0;
						E_POS <= 0;
						W_TOT <= 0;
						L_DER <= 0;
						E[0] <= 0;
						E[1] <= 0;
						Ld[0] <= 0;
						Ld[1] <= 0;		
						wt[0] <= 0;
						wt[1] <= 0;
						tm <= 0;
						errors <= 0;
						cnt <= 0;		
						e_cnt <= 0;		
						ctrl <= 0;				
					end
				endcase
			end
		end
endmodule
