//-----------------------------------------------------------------------------
//
// Title       : Modified_Error_corr
// Design      : decoder_5bit
// Author      : Sindre
// File        : Modified_Error_corr.v
//
//-----------------------------------------------------------------------------
//
// Description : Modified error correction module, use with modified Chien
//				 and Forney modules. Do to start to correct errors before
//				 all error positions and values are found.
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

//{module {Modified_Error_corr}}
module Modified_Error_corr #(parameter width=5)(
	input clk,							//clock signal
	input reset,						//reset signal
	input enable,						//enable signal
	input EP_ready,						//error pos. ready signal
	input EV_ready,						//error value ready
	input [width-1:0] E_value,			//error value
	input [width-1:0] POS,				//error position
	input [1:0] E_valid,				//valid errors
	input [width-1:0] S_IN,				//symbol input from codeword storage
	output reg RE,						//read enable
	output reg RS_DONE,					//RS decoding done
	output reg [width-1:0] add_read,	//RAM read address 
	output [width-1:0] S_OUT			//corrected symbol value
	);
	
	//internal constants
	parameter size = 3;
	parameter IDLE = 3'b000;
	parameter STEP1 = 3'b001;
	parameter STEP2 = 3'b011;
	parameter STEP3 = 3'b010;
	parameter STEP4 = 3'b110;
	
	//internal variables
	reg [1:0] errors;			//number of valid errors 
	reg [4:0] cnt;				//counter: max 31 - number of symbols
	reg [1:0] e_cnt;			//error counter
	reg [width-1:0] value;		//error value
	reg [width-1:0] pos [0:1];	//store error position
	reg [width-1:0] val [0:1];	//store error values
	reg [size-1:0] state;
	reg [size-1:0] next_state;
	reg gclk;					//clock gate signal
	reg qint;
	
	assign S_OUT = S_IN^value;	//finite field addition (correct error)
	
	//clock gating
	always@(clk or enable)
	begin
		if(clk == 0)begin
			qint <= (!enable);
		end
		gclk <= (!qint) && clk;
	end
	
	//FSM
	always @(posedge gclk or negedge reset)
	begin: FSM_SEQ
		if(reset == 0)begin
			state <= IDLE;
		end else if(enable)begin
			state <= next_state;	
		end
	end
	
	//change stat section
	always @(state or EP_ready or EV_ready or cnt or errors)
	begin: FSM_COMBO
		case(state)
			IDLE: begin
				if(EP_ready)begin
					next_state = STEP1;	
				end else begin
					next_state = IDLE;
				end
			end
			
			STEP1: begin
				if(cnt == errors-1)begin
					//when all error positions received
					//go to next state
					next_state = STEP2;	
				end else begin
					next_state = STEP1;
				end
			end
			
			STEP2: begin
				if(EV_ready)begin
					if(cnt == errors-1)begin
						//error values received
						//go to correction state
						next_state = STEP3;	
					end else begin
						next_state = STEP2;
					end
				end else begin
					next_state = STEP2;
				end
			end
			
			STEP3: begin
				if(cnt == 0)begin
					//correct last two symbols
					next_state = STEP4;
				end else begin
					next_state = STEP3;
				end
			end
	
			STEP4: begin
				if(cnt == 0)begin
					//all error corrected, go balc to IDLE state
					next_state = IDLE;
				end else begin
					next_state = STEP4;
				end
			end
			
			default: next_state = IDLE;
		endcase
	end
	
	//output section
	always @(posedge gclk or negedge reset)
	begin: FSM_OUT
		if(reset == 0)begin
		   //reset registers
		   RE <= 0;
		   cnt <= 0;
		   e_cnt <= 0;
		   errors <= 0;
		   RS_DONE <= 0;
		   add_read <= 0;
		   value <= 0;
		   pos[0] <= 0;
		   pos[1] <= 0;
		   val[0] <= 0;
		   val[1] <= 0;
		end else if(enable) begin
			case(state)
				IDLE: begin
					RE <= 0;
					cnt <= 0;
					RS_DONE <= 0;
					add_read <= 0;
					e_cnt <= 0;
					value <= 0;
					pos[0] <= 0;
		   			pos[1] <= 0;
		   			val[0] <= 0;
		   			val[1] <= 0;
					if(EP_ready)begin
						//error positions ready
						errors <= E_valid;	//number of error pos.
					end else begin
						errors <= 0;
					end
				end
				
				STEP1: begin
					//shift in error positions
					pos[cnt] <= POS;
					if(cnt == errors-1)begin
						cnt <= 0;	
					end else begin
						cnt <= cnt+1;
					end
				end
				
				STEP2: begin
					if(EV_ready)begin
						//error values ready
						//shift in error values
						val[cnt] <= E_value;
						if(cnt == errors-1)begin
							cnt <= 0;
							RE <= 1;
							cnt <= 29;
							add_read <= 30;	//read first symbol from memory
						end else begin
							cnt <= cnt+1;
							RE <= 0;
							add_read <= 0;
						end
					end else begin
						val[cnt] <= 0;
						cnt <= 0;
						RE <= 0;
						cnt <= 0;
						add_read <= 0;
					end
				end
				
				STEP3: begin
					//read in symbols and correct errors
					RE <= 1;
					add_read <= cnt;
					if(cnt+1 == pos[e_cnt])begin
						//if error at position, correct it
						value <= val[e_cnt];
						if(e_cnt == 1)begin
							e_cnt <= 0;
						end else begin
							e_cnt <= e_cnt+1;
						end
					end else begin
						value <= 0;
						e_cnt <= e_cnt; 
					end
					
					if(cnt == 0)begin
						cnt <= 0;
					end else begin
						cnt <= cnt-1;
					end
				
					if(cnt == 29)begin
						//corrected symbols ready
						RS_DONE <= 1;
					end else begin
						RS_DONE <= 0;
					end		
				end

				STEP4: begin
					//correct last two symbols
					if(cnt == pos[e_cnt])begin
						value <= val[e_cnt];
						if(e_cnt == 1)begin
							e_cnt <= 0;
						end else begin
							e_cnt <= e_cnt+1;
						end
					end else begin
						value <= 0;
						e_cnt <= e_cnt; 
					end

					if(cnt == 0)begin
						//count is zero, all symbols checked and corrected
						cnt <= 0;
					end else begin
						cnt <= 0;
					end		
				end
				
				default: begin
					RE <= 0;
					cnt <= 0;
					e_cnt <= 0;
					errors <= 0;
					RS_DONE <= 0;
					add_read <= 0;
					pos[0] <= 0;
					pos[1] <= 0;
					val[0] <= 0;
					val[1] <= 0;
				end
			endcase
		end
	end

endmodule
