//-----------------------------------------------------------------------------
//
// Title       : Modified_Forney
// Design      : decoder_5bit
// Author      : Sindre
// File        : Modified_Forney.v
//
//-----------------------------------------------------------------------------
//
// Description : Modified Forney's method. Calcualtes only error values for
//				 known error positions
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

//{module {Modified_Forney}}
module Modified_Forney #(parameter width=5)(
	input clk,						//clock signal
	input reset,					//reset signal
	input enable,					//enable signal
	input EP_ready,					//error position ready
	input [1:0] E_valid,			//valid errors
	input [width-1:0] E_POS,		//error position
	input [width-1:0] W_TOT,		//omega(x) values
	input [width-1:0] L_DER,    	//error locator polynomial derived
	input [width-1:0] INV,			//inverse value
	output reg [width-1:0] VAL,		//value to find inverse of
	output reg RE,					//sread enable signal for INVERSE ROM
	output reg EV_ready,	    	//error value ready signal
	output reg [width-1:0] E_value	//error value
	);
	
	//internal constants
	parameter size = 3;
	parameter IDLE = 3'b000;
	parameter STEP1 = 3'b001;
	parameter STEP2 = 3'b011;
	parameter STEP3 = 3'b010;
	
	//internal variables
	reg [4:0] cnt;					//counter: max 31 - number of symbols
	reg [1:0] errors;				//number of errors
	reg [width-1:0] values [0:1];	//error values
	reg [size-1:0] state;
	reg [size-1:0] next_state;
	reg gclk;
	reg qint;
	
	//internal wires and reg
	wire [width-1:0] M_OUT;		//output of	full multiplier
	wire [width-1:0] NOR_OUT;	
	reg [width-1:0] R_OUT01;	//delay registers while inverse value are found
	reg [width-1:0] R_OUT02;	//delay registers
	reg [width-1:0] R_OUT03;	//delay registers
	reg [width-1:0] R_OUT04;	//delay registers
	reg [width-1:0] R_OUT05;	//delay registers
	reg [width-1:0] R_OUT06;	//delay registers
	reg [width-1:0] R_OUT07;	//delay registers
	wire [width-1:0] VAL_OUT;	//output value
	
	//module instantiations
	assign VAL_OUT = R_OUT06 & R_OUT07;	 //AND the results
	
	//clock gating
	always@(clk or enable)
	begin
		if(clk == 0)begin
			qint <= (!enable);
		end
		gclk <= (!qint) && clk;
	end
	
	//finite field multiplication
	Full_mult #(.width(width)) FFM01(
	.A	(R_OUT04),
	.B	(INV),
	.Y	(M_OUT)
	);
	
	//FSM
	always @ (posedge gclk or negedge reset)
		begin: FSM_SEQ
			if (reset == 0)begin
				state <= IDLE;
			end else if(enable) begin
				state <= next_state;
			end
		end
	//change state section
	always @ (state or EP_ready or cnt or errors)
		begin: FSM_COMBO
			case(state)
				IDLE: begin
					if(EP_ready)begin
						//if error positions ready start shift in error positions
						next_state = STEP1;
					end else begin
						next_state = IDLE;
					end 
				end
				
				STEP1: begin
					//shift in shift error position
					next_state = STEP2;	
				end
				
				STEP2: begin
					//calcualte error values
					if(cnt == errors)begin
						next_state = STEP3;
					end else begin
						//shift in second erro position if any
						next_state = STEP2;
					end
				end
				
				//shift out error values
				STEP3: begin
					if(cnt == errors)begin
						//go back to IDLE when all values shifted out
						next_state = IDLE;
					end else begin
						next_state = STEP3;
					end
				end
				
				default: next_state = IDLE; 
			endcase
		end	
	
	always @ (posedge gclk or negedge reset)
		begin: FSM_OUT
			if(reset == 0)begin
				//reset registers
				errors <= 0;
				cnt <= 0;
				RE <= 0;				
				EV_ready <= 0;
				values[0] <= 0;
				values[1] <= 0;
				R_OUT01 <= 0;
				R_OUT02 <= 0;
				R_OUT03 <= 0;
				R_OUT04 <= 0;
				R_OUT05 <= 0;
				R_OUT06 <= 0;
				R_OUT07 <= 0;
			end else if(enable) begin
				case(state)
					IDLE: begin
						EV_ready <= 0;
						RE <= 0;
						cnt <= 0;
						errors <= 0;
					end
					
					STEP1: begin
						RE <= 1;			//read select set high to read from ROM
						cnt <= 0;
						errors <= E_valid;	//number of valid errors
						EV_ready <= 0;
						R_OUT01 <= W_TOT;	//input omega(x) value
						R_OUT04 <= R_OUT01;
						VAL <= L_DER;		//finde inverse value of locator poly derived value
						R_OUT06 <= M_OUT; 
						R_OUT03 <= E_POS;	//input error pos
						R_OUT05 <= R_OUT03;
						R_OUT07 <= R_OUT05;
					end
					
					STEP2: begin
						//shift in positions and calcualte values
						R_OUT01 <= W_TOT;
						R_OUT04 <= R_OUT01;
						VAL <= L_DER;
						R_OUT06 <= M_OUT; 
						R_OUT03 <= E_POS;
						R_OUT05 <= R_OUT03;
						R_OUT07 <= R_OUT05;
						if(VAL_OUT != 0)begin
							if(cnt == errors)begin
								//store error value
								RE <= 0;
								values[cnt] <= VAL_OUT;
								cnt <= 0;
							end else begin
								//store error value
								RE <= 1;
								values[cnt] <= VAL_OUT;
								cnt <= cnt+1;
							end
						end else begin
							RE <= 1;
							cnt <= 0;
							values[cnt] <= 0;
						end
					end
					
					//shift out error values
					STEP3: begin
						if(cnt == errors)begin
							//all values shifted out
							EV_ready <= 0;
							E_value <= 0;
							cnt <= 0;
						end else begin
							EV_ready <= 1; //error values ready
							E_value <= values[cnt];	//shift out error values
							cnt <= cnt+1;
						end
					end
					
					default: begin
						errors <= 0;
						RE <= 0;				
						EV_ready <= 0;
						cnt <= 0;
						values[0] <= 0;
						values[1] <= 0;
					end
				endcase
			end
		end

endmodule
