//-----------------------------------------------------------------------------
//
// Title       : decoder_top
// Design      : decoder_5bit
// Author      : Sindre
// File        : decoder_top.v
//
//-----------------------------------------------------------------------------
//
// Description : 5-bit RS decoder top module: Configuration 6
//				 Syndrome module uses two parallel syndromes cells.
//				 Uses a two ports data output memory module that enables parallel
//				 reading from memory.
//				 
//				 
//				 This decoder computes all values in Chien search module first
//				 (i.e no pipelining). Then only valid error positions are
//				 computed using Forney algorithm (max 2 values)
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

//{module {decoder_top}}
module decoder_top #(parameter width=5)(
	input clk,						//clock signal
	input reset,					//reset signal
	input enable,					//enable signal 
	input rs_start,				   	//RS decoding start signal
	input [width-1:0] data_in,		//symbol data input
	output NO_ERRORS,				//no errors detected signal, all syndromes are zero
	output NO_COR,					//high value indicates to many errors to correct
	output rs_done,					//RS decoding done
	output [width-1:0] data_out		//corrected symbol output
	);
	
	//internal regs
	reg sel_1;				//select signal memory selection

	//enable signals
	reg en1;		//enable signal for in stage module
	reg en2;		//enable signal for syndrome module
	reg en3;		//enable signal for Berlekamp-Massey module 
	reg en4;		//enable signal for error locator calculation module
	reg en5;		//enable signal for Chien search
	reg en6;		//enable signal for Forney moduel
	reg en7;		//enable signal for error correction module
	
	//internal wires
	wire WE_MEM01;
	wire RE1_MEM01;
	wire RE2_MEM01;
	
	wire NOT_VALID;
	wire RE_INV;
	wire i_start;
	wire s_ready;
	wire l_ready;
	wire lw_ready;
	wire ep_ready;
	wire ev_ready;

	
	wire [width-1:0] MUX_ADD1;
	wire [width-1:0] ADD_WRITE;
	wire [width-1:0] ADD_READ1;
	wire [width-1:0] ADD_READ2;
	wire [width-1:0] ADD_READ3;
	wire [width-1:0] IN_STAGE_OUT;
	wire [width-1:0] MEM01_OUT1;
	wire [width-1:0] MEM01_OUT2;
	wire [width-1:0] MUX_MEM_OUT1;
	wire [width-1:0] MUX_MEM_OUT2;
	wire [width-1:0] S1;
	wire [width-1:0] S2;
	wire [width-1:0] S3;
	wire [width-1:0] S4;
	wire [width-1:0] L0;
	wire [width-1:0] L1;
	wire [width-1:0] L2;
	wire [width-1:0] W0;
	wire [width-1:0] W1;
	wire [width-1:0] error_pos;
	wire [width-1:0] error_val;
	wire [width-1:0] w_tot;
	wire [width-1:0] l_der;
	wire [width-1:0] pos;
	wire [1:0] valid_errors;
	wire [width-1:0] ADD_READ_INV;
	wire [width-1:0] INV_OUT;

	reg RS_START;
	reg I_START;
	reg S_READY;
	reg L_READY;
	reg LW_READY;
	reg EP_READY;
	reg EV_READY;
	reg [width-1:0] DATA_IN;
	reg [width-1:0] ERROR_POS;
	reg [width-1:0] W_TOT;
	reg [width-1:0] L_DER;
	reg [width-1:0] ERROR_VAL;
	reg [width-1:0] POS;
	reg [1:0] VALID_ERRORS;


	//assigns
	assign NO_COR = NOT_VALID;
	
	//multiplex control for reading from memory 1
	assign MUX_ADD1 = sel_1? ADD_READ3 : ADD_READ1;
	
	//-----------------------//
	// Module instantiations //
	//-----------------------//
	
	//RS decoder input stage
	in_stage #(.width(width)) INPUT_STAGE(
	.clk			(clk),			//clock signal
	.reset			(reset),		//reset signal
	.enable			(en1),			//enable signal
	.rs_start		(RS_START),		//start decoding
	.data_in		(DATA_IN),		//symbol input
	.i_start		(i_start),		//start syndrome calculations signal
	.WE				(WE_MEM01),		//write enable
	.address_write	(ADD_WRITE),	//write address
	.data_out		(IN_STAGE_OUT)	 //output bus for in_stage
	);
	
	//Codeword memory storage
	Codeword_Store_modified #(.width(width)) MEM01(
	.clk			(clk),			//clock signal
	.enable			(enable),		//enable signal
	.we				(WE_MEM01),		//write enable
	.re1			(RE1_MEM01),	//read enable (parallel read)
	.re2			(RE2_MEM01),	//read enable 
	.address_write	(ADD_WRITE),  	//write address
	.address_read1	(MUX_ADD1),  	//read address 1 (no parallel mode uses address_read1)
	.address_read2	(ADD_READ2),  	//read address 2
	.data_in		(IN_STAGE_OUT),	//data input
	.data_out1		(MEM01_OUT1), 	//data output 1 (no parallel mode uses data_out1)
	.data_out2		(MEM01_OUT2)  	//data output 2
	);
	
	//MUX for controlling memory output 1
	MUX_1to2 #(.width(width)) MUX_MEM_OUT(
	.in		(MEM01_OUT1),  	//mux input
	.sel	(sel_1), 		//select signal
	.out1	(MUX_MEM_OUT1),	//output 1 connected to syndrome module
	.out2	(MUX_MEM_OUT2)	//output 2 connected to error corr. module
	);
	
	//Syndrome calculator
	Modified_syndrome_ctrl #(.width(width)) SYNDROME(
	.clk			(clk),			//clock signal
	.reset			(reset),		//reset signal
	.enable			(en2),			//enable signal
	.S_START		(I_START),		//Syndrome computation start signal
	.s_in1			(MUX_MEM_OUT1),	//input 1 from codeword memory
	.s_in2			(MEM01_OUT2),	//input 2 from codeword memory
	.RE				(RE1_MEM01),	//read enable, reading from codeword memory
	.S_ready		(s_ready),		//syndrome values ready
	.NO_ERRORS		(NO_ERRORS),	//no errors flag
	.address_read1	(ADD_READ3),	//memory address read 1
	.address_read2	(ADD_READ2),	//memory address read 2
	.S1				(S1),			//syndrome value 1
	.S2				(S2),			//syndrome value 2
	.S3				(S3),			//syndrome value 3
	.S4				(S4)			//syndrome value 4
	);
	
	//inversionless Berlekamp-Massey algorithm
	//calculates error locator polynomial
	BM_alg #(.width(width)) BM_Lamda(
	.clk		(clk),		//clock signal
	.reset		(reset),	//reset signal
	.enable		(en3),		//enable signal
	.S_ready	(S_READY),	//syndromes ready signal
	.S1			(S1),		//syndrome value 1
	.S2			(S2),		//syndrome value 2
	.S3			(S3),		//syndrome value 3
	.S4			(S4),		//syndrome value 4
	.L_ready	(l_ready),	//error locator poly. ready signal
	.L0			(L0),		//locator poly. coefficient 0 output
	.L1			(L1),		//locator poly. coefficient 1 output
	.L2			(L2)		//locator poly. coefficient 2 output
	);
	
	//Error evaluator polynomial calculation
	Omega_calc #(.width(width)) OMEGA(
	.clk		(clk),		//clock signal
	.reset		(reset),	//reset signal
	.enable		(en4),		//enable signal
	.L_ready	(L_READY),	//error locator poly. ready signal
	.S1			(S1),		//syndrome value 1
	.S2			(S2),		//syndrome value 2
	.S3			(S3),		//syndrome value 3
	.S4			(S4),		//syndrome value 4
	.L0			(L0),		//locator poly. coefficient 0 input
	.L1			(L1),		//locator poly. coefficient 1 input
	.L2			(L2),		//locator poly. coefficient 2 input
	.LW_ready	(lw_ready),	//error evaluator poly. ready signal
	.W0			(W0),		//evaluator poly. coefficient 0 output
	.W1			(W1)		//evaluator poly. coefficient 1 output
	);
	
	//Chien search
	Modified_Chien_search #(.width(width)) CHIEN( 
	.clk		(clk),			//clock signal
	.reset		(reset),		//reset signal
	.enable		(en5),			//enable signal
	.LW_ready	(LW_READY),		//error evaluator poly. ready signal
	.L0			(L0),			//locator poly. coefficient 0 input
	.L1			(L1),			//locator poly. coefficient 1 input
	.L2			(L2),			//locator poly. coefficient 2 input
	.W0			(W0),			//evaluator poly. coefficient 0 input
	.W1			(W1),			//evaluator poly. coefficient 1 input
	.NO_COR		(NOT_VALID),	//not valid signal
	.EP_ready	(ep_ready),		//error position ready
	.E_valid	(valid_errors),	//number of valid errors
	.POS		(pos),			//real error position output
	.E_POS		(error_pos),	//error position
	.W_TOT		(w_tot),		//Omega(x) value
	.L_DER		(l_der)			//error locator polynomial derived
	);
	
	//Forney algorithm
	Modified_Forney #(.width(width)) FORNEY(
	.clk		(clk),			//clock signal
	.reset		(reset),		//reset signal
	.enable		(en6),			//enable signal
	.EP_ready	(EP_READY),		//error position ready
	.E_valid	(VALID_ERRORS),	//valid errors
	.E_POS		(ERROR_POS),	//error position
	.W_TOT		(W_TOT),		//omega(x) values
	.L_DER		(L_DER),    	//error locator polynomial derived
	.INV		(INV_OUT),		//inverse value input
	.VAL		(ADD_READ_INV),	//value to find inverse of
	.RE			(RE_INV),		//read enable
	.EV_ready	(ev_ready),	    //error value ready signal
	.E_value	(error_val)		//error value
	);

	//ROM storing inverse values of GF(32)
	inverse_GF32 #(.width(width)) INVERSEROM(
	.clk			(clk),			//clock signal
	.re				(RE_INV),		//read enable
	.enable			(enable),		//enable signal
	.address_read	(ADD_READ_INV), //read address
	.data_out		(INV_OUT)		//data output
	);

	//Error correction
	Modified_Error_corr #(.width(width)) CORRECTION(
	.clk		(clk),				//clock signal
	.reset		(reset),			//reset signal
	.enable		(en7),				//enable signal used for CG
	.EP_ready	(EP_READY),			//error pos. ready signal
	.EV_ready	(EV_READY),			//error value ready	signal
	.E_value	(ERROR_VAL),		//error value
	.POS		(POS),				//error position
	.E_valid	(VALID_ERRORS),		//valid errors
	.S_IN		(MUX_MEM_OUT2),		//symbol input from codeword storage
	.RE			(RE2_MEM01),		//read enable (memory module)
	.RS_DONE	(rs_done),			//RS decoding done
	.add_read	(ADD_READ1),		//RAM read address 
	.S_OUT		(data_out)			//corrected symbol value
	);
	
	//controlling select 1 signal based module control signals
	//select 1 is used in MUX module
	always @(posedge clk or negedge reset)
	begin
		if(!reset)begin
			sel_1 <= 0;	
		end else if(enable)begin
			if(rs_start)begin
				sel_1 <= 1;
			end else if(s_ready)begin
				sel_1 <= 0;
			end else begin
				sel_1 <= sel_1;
			end 
		end 
	end

	//select which modules to enabled
	always @(posedge clk or negedge reset)
	begin: MODULE_ENABLE
		if(!reset)begin
			en1 <= 0;
			en2 <= 0;
			en3 <= 0;
			en4 <= 0;
			en5 <= 0;
			en6 <= 0;
			en7 <= 0;
		end else if(enable)begin
			if(rs_start)begin
				en1 <= 1;
				en2 <= 0;
				en3 <= 0;
				en4 <= 0;
				en5 <= 0;
				en6 <= 0;
				en7 <= 0;
			end else if(i_start)begin
				en1 <= 1;
				en2 <= 1;
				en3 <= 0;
				en4 <= 0;
				en5 <= 0;
				en6 <= 0;
				en7 <= 0;
			end else if(s_ready)begin
				en1 <= 0;
				en2 <= 0;
				en3 <= 1;
				en4 <= 0;
				en5 <= 0;
				en6 <= 0;
				en7 <= 0;				
			end else if(l_ready)begin
				en1 <= 0;
				en2 <= 0;
				en3 <= 0;
				en4 <= 1;
				en5 <= 0;
				en6 <= 0;
				en7 <= 0;			
			end else if(lw_ready)begin
				en1 <= 0;
				en2 <= 0;
				en3 <= 0;
				en4 <= 0;
				en5 <= 1;
				en6 <= 0;
				en7 <= 0;				
			end else if(ep_ready)begin
				en1 <= 0;
				en2 <= 0;
				en3 <= 0;
				en4 <= 0;
				en5 <= 1;
				en6 <= 1;
				en7 <= 1;				
			end else if(ev_ready)begin
				en1 <= 0;
				en2 <= 0;
				en3 <= 0;
				en4 <= 0;
				en5 <= 0;
				en6 <= 1;
				en7 <= 1;
			end else if(rs_done)begin
				en1 <= 0;
				en2 <= 0;
				en3 <= 0;
				en4 <= 0;
				en5 <= 0;
				en6 <= 0;
				en7 <= 1;	
			end else begin
				en1 <= en1;
				en2 <= en2;
				en3 <= en3;
				en4 <= en4;
				en5 <= en5;
				en6 <= en6;
				en7 <= en7;			
			end
		end else if(!enable)begin
			en1 <= 0;
			en2 <= 0;
			en3 <= 0;
			en4 <= 0;
			en5 <= 0;
			en6 <= 0;
			en7 <= 0;
		end
	end	
   
	//delay all bus values 1 cc, when using enable signal to enable modules
	always @(posedge clk or negedge reset)
	begin
		if(!reset)begin
			DATA_IN <= 0;
			ERROR_POS <= 0;
			ERROR_VAL <= 0;
			VALID_ERRORS <= 0;
			POS <= 0;
			W_TOT <= 0;
			L_DER <= 0;
			RS_START <= 0;
			S_READY <= 0;
			L_READY <= 0;
			LW_READY <= 0;
			EP_READY <= 0;
			EV_READY <= 0;
		end else if(enable) begin
			if(rs_start)begin
				DATA_IN <= data_in;
				ERROR_POS <= error_pos;
				ERROR_VAL <= error_val;
				VALID_ERRORS <= valid_errors;
				POS <= pos;
				W_TOT <= w_tot;
				L_DER <= l_der;
				RS_START <= 1;
				I_START <= 0;
				S_READY <= 0;
				L_READY <= 0;
				LW_READY <= 0;
				EP_READY <= 0;
				EV_READY <= 0;
			end else if(i_start) begin
				DATA_IN <= data_in;
				ERROR_POS <= error_pos;
				ERROR_VAL <= error_val;
				VALID_ERRORS <= valid_errors;
				POS <= pos;
				W_TOT <= w_tot;
				L_DER <= l_der;
				RS_START <= 0;
				I_START <= 1;
				S_READY <= 0;
				L_READY <= 0;
				LW_READY <= 0;
				EP_READY <= 0;
				EV_READY <= 0;
			end else if(s_ready)begin
				DATA_IN <= data_in;
				ERROR_POS <= error_pos;
				ERROR_VAL <= error_val;
				VALID_ERRORS <= valid_errors;
				POS <= pos;
				W_TOT <= w_tot;
				L_DER <= l_der;
				RS_START <= 0;
				I_START <= 0;
				S_READY <= 1;
				L_READY <= 0;
				LW_READY <= 0;
				EP_READY <= 0;
				EV_READY <= 0;
			end else if(l_ready)begin
				DATA_IN <= data_in;
				ERROR_POS <= error_pos;
				ERROR_VAL <= error_val;
				VALID_ERRORS <= valid_errors;
				POS <= pos;
				W_TOT <= w_tot;
				L_DER <= l_der;
				RS_START <= 0;
				S_READY <= 0;
				I_START <= 0;
				L_READY <= 1;
				LW_READY <= 0;
				EP_READY <= 0;
				EV_READY <= 0;
			end else if(lw_ready)begin
				DATA_IN <= data_in;
				ERROR_POS <= error_pos;
				ERROR_VAL <= error_val;
				VALID_ERRORS <= valid_errors;
				POS <= pos;
				W_TOT <= w_tot;
				L_DER <= l_der;
				RS_START <= 0;
				I_START <= 0;
				S_READY <= 0;
				L_READY <= 0;
				LW_READY <= 1;
				EP_READY <= 0;
				EV_READY <= 0;
			end else if(ep_ready)begin
				DATA_IN <= data_in;
				ERROR_POS <= error_pos;
				ERROR_VAL <= error_val;
				VALID_ERRORS <= valid_errors;
				POS <= pos;
				W_TOT <= w_tot;
				L_DER <= l_der;
				RS_START <= 0;
				I_START <= 0;
				S_READY <= 0;
				L_READY <= 0;
				LW_READY <= 0;
				EP_READY <= 1;
				EV_READY <= 0;
			end else if(ev_ready)begin
				DATA_IN <= data_in;
				ERROR_POS <= error_pos;
				ERROR_VAL <= error_val;
				VALID_ERRORS <= valid_errors;
				POS <= pos;
				W_TOT <= w_tot;
				L_DER <= l_der;
				RS_START <= 0;
				I_START <= 0;
				S_READY <= 0;
				L_READY <= 0;
				LW_READY <= 0;
				EP_READY <= 0;
				EV_READY <= 1;
			end else begin
				DATA_IN <= data_in;
				ERROR_POS <= error_pos;
				ERROR_VAL <= error_val;
				VALID_ERRORS <= valid_errors;
				POS <= pos;
				W_TOT <= w_tot;
				L_DER <= l_der;
				RS_START <= 0;
				I_START <= 0;
				S_READY <= 0;
				L_READY <= 0;
				LW_READY <= 0;
				EP_READY <= 0;
				EV_READY <= 0;
			end
		end
	end
endmodule

	

	
