//-----------------------------------------------------------------------------
//
// Title       : Chien_search
// Design      : decoder_5bit
// Author      : Sindre
// File        : Chien_search.v
//
//-----------------------------------------------------------------------------
//
// Description : Chien search module
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

//{module {Chien_search}}
module Chien_search #(parameter width=5) (
	input clk,				//clock signal
	input reset,		    //reset signal
	input enable,			//enable signal
	input LW_ready,			//error evaluator poly. ready signal
	input [width-1:0] L0,	//locator poly. coefficient 0 input
	input [width-1:0] L1,	//locator poly. coefficient 1 input
	input [width-1:0] L2,	//locator poly. coefficient 2 input
	input [width-1:0] W0,	//evaluator poly. coefficient 0 input
	input [width-1:0] W1,	//evaluator poly. coefficient 1 input
	output reg EP_ready,		//error position ready signal
	output reg E_valid,			//valid errors
	output [width-1:0] E_POS,	//error position 
	output [width-1:0] W_TOT,	//outputs the calculated omega(x) value at each clock cycle
	output [width-1:0] L_DIR	///error locator polynomial derived
	);
	
	//internal constants
	parameter A0 = 5'b00001;	//decimal value 1
	parameter A1 = 5'b00010;	//decimal value 2
	parameter A2 = 5'b00100;	//decimal value 4
	
	parameter size = 3;
	parameter IDLE = 3'b000;
	parameter STEP1 = 3'b001;
	parameter STEP2 = 3'b011;
	
	//internal variables
	reg [width-1:0] E[0:1];		//internal storage for error positions
	reg [width-1:0] cnt;		//counter: max 31 - number of symbols
	reg [size-1:0] state;
	reg [size-1:0] next_state;
	reg [1:0] e_cnt;			//counter used for shifting out error values
	reg [1:0] errors;			//counter to track number of error positions, max 2
	reg ctrl;					//Control signal for selecting input on multiplexers
	reg tm;						//too many errors flag, if set high indicates that there are more the two errors
	reg gclk;
	reg qint;
		
		
	//internal wires
	wire [width-1:0] OUT_CELL00;
	wire [width-1:0] OUT_CELL01;
	wire [width-1:0] OUT_CELL02;
	wire [width-1:0] OUT_CELL03;
	wire [width-1:0] OUT_CELL04;
	wire [width-1:0] OUT_FFA01;
	wire [width-1:0] OUT_FFA02;
	wire [width-1:0] OUT_FFA03;
	
	assign L_DIR = OUT_CELL01;
	assign W_TOT = OUT_FFA03;
	
	//-------------------------//
	//  Module instantiations  //
	//-------------------------//
	
	//calculation of lamda(x) values
	Chien_cell #(.width(width)) CC00(
	.clk	(gclk),		 //clock signal
	.reset	(reset),	 //reset signal
	.enable	(enable),	 //enable signal
	.sel	(ctrl),		 //select signal
	.L		(L0),		 //error locator coefficient 0 input
	.a		(A0),		 //generator root alpha^0 input
	.out	(OUT_CELL00) //output
	);
	
	Chien_cell #(.width(width)) CC01(
	.clk	(gclk),		 //clock signal
	.reset	(reset),	 //reset signal
	.enable	(enable),	 //enable signal
	.sel	(ctrl),		 //select signal
	.L		(L1),		 //error locator coefficient 1 input
	.a		(A1),		 //generator root alpha^1 input
	.out	(OUT_CELL01) //output
	);
	
	Chien_cell #(.width(width)) CC02(
	.clk	(gclk),		//clock signal
	.reset	(reset),	//reset signal
	.enable	(enable),	//enable signal
	.sel	(ctrl),		//select signal
	.L		(L2),		//error locator coefficient 2 input
	.a		(A2),		//generator root alpha^2 input
	.out	(OUT_CELL02)//output
	);
	
	//add the output from cells
	Adder #(.width(width)) FFA01(
	.in1	(OUT_CELL00),
	.in2	(OUT_CELL01),
	.sum	(OUT_FFA01)
	);
	
	Adder #(.width(width)) FFA02(
	.in1	(OUT_FFA01),
	.in2	(OUT_CELL02),
	.sum	(OUT_FFA02)
	);
	
	//calculation of omega(x) values
	Chien_cell_W #(.width(width)) CCW00(
	.clk	(gclk),		//clock signal
	.reset	(reset),	//reset signal
	.enable	(enable),	//enable signal
	.sel	(ctrl),		//select signal
	.W		(W0),		//error evaluator coefficient 0 input
	.a		(A0),		//generator root alpha^0 input
	.out	(OUT_CELL03)//output
	);
	
	Chien_cell_W #(.width(width)) CCW01(
	.clk	(gclk),		 //clock signal
	.reset	(reset),	 //resey signal
	.enable	(enable),	 //enable signal
	.sel	(ctrl),		 //select signal
	.W		(W1),		 //error evaluator coefficient 0 input
	.a		(A1),		 //generator root alpha^1 input
	.out	(OUT_CELL04) //output
	);
	
	Adder #(.width(width)) FFA03(
	.in1	(OUT_CELL03),
	.in2	(OUT_CELL04),
	.sum	(OUT_FFA03)
	);
	
	NOR NOR01(
	.IN		(OUT_FFA02),
	.OUT	(E_POS)
	);
	
	//clock gating
	always@(clk or enable)
	begin
		if(clk == 0)begin
			qint <= (!enable);
		end
		
		gclk <= (!qint) && clk;
	end
	
	//FSM
	always @ (posedge gclk or negedge reset)
		begin: FSM_SEQ
			if (reset == 0)begin
				state <= IDLE;
			end else if(enable) begin
				state <= next_state;
			end
		end
	
	//change state section
	always @ (state or LW_ready or cnt or tm)
		begin: FSM_COMBO
			case(state)
				IDLE: begin
					if(LW_ready == 1) begin
						//start calculations when polynomials are ready 
						next_state = STEP1;
					end else begin
						next_state = IDLE;
					end	
				end
						
				STEP1: begin
					if(tm == 1)begin
						next_state = IDLE;
					end else begin
						if(cnt == 0) begin
							//go back to IDLE when all positions have been checked
							next_state = IDLE;
						end else begin
							next_state = STEP1;
						end
					end
				end
							
				default: next_state = IDLE; 
			endcase
		end	
	
	//output section
	always @ (posedge gclk or negedge reset)
		begin: FSM_OUT
			if(reset == 0)begin
				//reset all registers and signals
				E_valid <= 0;
				EP_ready <= 0;
				tm <= 0;
				E[0] <= 0;
				E[1] <= 0;
				errors <= 0;	//Number of errors set to zero
				cnt <= 0;		//Reset counter to zero
				e_cnt <= 0;		//Reset counter
				ctrl <= 0;		//Lamda and omega coeff is selected on the MUX in Chien cell
			end else if(enable) begin
				case(state)
					IDLE: begin
						E_valid <= 0;
						EP_ready <= 0;
						E[0] <= 0;
						E[1] <= 0;
						errors <= 0;
						e_cnt <= 0;
						ctrl <= 0;
						EP_ready <= 0;
						if(LW_ready) begin
							//error locator and evaluator polys. ready
							cnt <= 30;	   //check position 30 for error
							EP_ready <= 1; //error position ready
							ctrl <= 1;
						end else begin
							cnt <= 0;
							EP_ready <= 0;
						end	
					end
					
					//check through all symbol positions for error
					STEP1: begin
						if(cnt == 0)begin
							ctrl <= 0;
							cnt <= 0;
						end else begin
						   EP_ready <= 0;
							ctrl <= 1;
							cnt <= cnt-1;
						end 
						//check if value is zero
						if(OUT_FFA02 == 0) begin
							//if zero: there is an error that position
							if(errors == 2)begin
								tm <= 1;
								E_valid <= 1;
							end else begin
								tm <= 0;
								E_valid <= 0;
								E[errors] <= cnt;
								errors <= errors+1;
							end
						end else begin
							E[errors] <= 0;
							errors <= errors;
						end
						
					end
						
					default: begin
						E_valid <= 0;
						EP_ready <= 0;
						E[0] <= 0;
						E[1] <= 0;
						errors <= 0;	
						cnt <= 0;		
						e_cnt <= 0;	
						ctrl <= 0;		
					end
				endcase
			end
		end
endmodule
