//-----------------------------------------------------------------------------
//
// Title       : Modified_syndrome_ctrl
// Design      : decoder_5bit
// Author      : Sindre
// File        : Modified_syndrome_ctrl.v
//
//-----------------------------------------------------------------------------
//
// Description : Modified syndrome computation:
//				 Implemented with two-parallel syndrome cells
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

//{module {Modified_syndrome_ctrl}}
module Modified_syndrome_ctrl #(parameter width=5)(
	input wire clk,					   	//clock signal
	input wire reset,				    //reset signal
	input wire enable,					//enable signal
	input wire S_START,					//Syndrome computation start signal
	input wire [width-1:0] s_in1,		//symbol input 1 
	input wire [width-1:0] s_in2,		//symbol input 2 
	output reg S_ready,					//syndrome values ready
	output reg NO_ERRORS,				//no errors flag
	output reg [width-1:0] S1,			//syndrome value 1
	output reg [width-1:0] S2,			//syndrome value 2
	output reg [width-1:0] S3,			//syndrome value 3
	output reg [width-1:0] S4			//syndrome value 4
	);
	
	//internal constants
	parameter size = 3;
	parameter IDLE = 3'b000;
	parameter STEP1 = 3'b001;
	parameter STEP2 = 3'b011;
	
	//internal variables
	reg control;
	reg [width-1:0] cnt;	//counter max value 31
	reg [4:0] wait_cnt;		//counter max value 16
	
	reg [size-1:0] state;
	reg [size-1:0] next_state;	
	reg gclk;
	reg qint;
	
	//internal wires
	wire [width-1:0] S_1;
	wire [width-1:0] S_2;
	wire [width-1:0] S_3;
	wire [width-1:0] S_4;
	reg [width-1:0] A1_IN;
	reg [width-1:0] B1_IN;
	reg [width-1:0] temp;
	
	//module instantiations
	Syndrome_cell_parallel_S0 #(.width(width)) SC1(
	.clk		(gclk),			//clock signal
	.reset		(reset),		//reset signal
	.enable		(enable),		//enable signal
	.ctrl		(control),		//control signal
	.in_A		(A1_IN),		//input A for even symbols
	.in_B		(B1_IN),		//input B for odd symbols
	.out		(S_1)			//output symdrome value
	);
	
	Syndrome_cell_parallel_S1 #(.width(width)) SC2(
	.clk		(gclk),			//clock signal
	.reset		(reset),		//reset signal
	.enable		(enable),		//enable signal
	.ctrl		(control),		//control signal
	.in_A		(A1_IN),		//input A for even symbols
	.in_B		(B1_IN),		//input B for odd symbols
	.out		(S_2)			//output symdrome value
	);
	
	Syndrome_cell_parallel_S2 #(.width(width)) SC3(
	.clk		(gclk),			//clock signal
	.reset		(reset),		//reset signal
	.enable		(enable),		//enable signal
	.ctrl		(control),		//control signal
	.in_A		(A1_IN),		//input A for even symbols
	.in_B		(B1_IN),		//input B for odd symbols
	.out		(S_3)			//output symdrome value
	);
	
	Syndrome_cell_parallel_S3 #(.width(width)) SC4(
	.clk		(gclk),			//clock signal
	.reset		(reset),		//reset signal
	.enable		(enable),		//enable signal
	.ctrl		(control),		//control signal
	.in_A		(A1_IN),		//input A for even symbols
	.in_B		(B1_IN),		//input B for odd symbols
	.out		(S_4)			//output symdrome value
	);
	
	//clock gating
	always@(clk or enable)
	begin
		if(clk == 0)begin
			qint <= (!enable);
		end
		
		gclk <= (!qint) && clk;
	end
	
	//FSM
	always @(posedge gclk or negedge reset)
		begin: FSM_SEQ
			if (reset == 0) begin
				state <= IDLE;
			end else if(enable) begin
				state <= next_state;
			end
		end
	
	//change state section
	always @(state or S_START or wait_cnt or cnt)
	begin: FSM_COMBO
		case(state)
			IDLE: begin
				if(S_START == 1) begin
					//start syndrome calculations
					next_state = STEP1;
				end else begin
					next_state = IDLE;
				end	
			end
			
			STEP1: begin
				if(cnt == 0)begin
					//go back to IDLE when calulations are done
					next_state = IDLE;
				end else begin
					next_state = STEP1;
				end
			end
			
			default: next_state = IDLE;		
		endcase
	end
	
	//output section
	always @(posedge gclk or negedge reset)
		begin: FSM_OUT
			if(reset == 0)begin
				//reset registers and counters
				wait_cnt <= 0;
				cnt <= 0;
				control <= 0;
				A1_IN <= 0;
				B1_IN <= 0;
				S_ready <= 0;
				S1 <= 0;
				S2 <= 0;
				S3 <= 0;
				S4 <= 0;
				NO_ERRORS <= 0;
			end else if(enable) begin
				case(state)
					IDLE: begin
						wait_cnt <= 0;
						cnt <= 0;
						control <= 0;
						A1_IN <= 0;
						B1_IN <= 0;
						S_ready <= 0;
						NO_ERRORS <= 0;
						if(S_START)begin
							//start syndrome calc.
							cnt <= 16;
							A1_IN <= s_in1;	//input first symbol
							B1_IN <= 0;
							temp <= s_in2;	//temporary store first odd symbol
						end else begin
							cnt <= 0;
							A1_IN <= 0;
							B1_IN <= 0;
							temp <= 0;
						end
					end
					
					STEP1: begin
						if(cnt == 0)begin
							if(S_1 == 0 && S_2 == 0 && S_3 == 0 && S_4 == 0)begin
								//if all syndrome values zero, no error
								S_ready <= 0;
								cnt <= 0;
								control <= 0;
								S1 <= 0;
								S2 <= 0;
								S3 <= 0;
								S4 <= 0;
								NO_ERRORS <= 1;
							end else begin
								//if not zero shift out symbol values
								S_ready <= 1;  //syndromes ready
								cnt <= 0;
								control <= 0;
								S1 <= S_1;
								S2 <= S_2;
								S3 <= S_3;
								S4 <= S_4;
								NO_ERRORS <= 0;
							end
						end else begin
							//calculate syndrome values
							A1_IN <= s_in1; //input even symbol
							temp <= s_in2;	//input odd symbol
							B1_IN <= temp;  //temp store odd symbol
							control <= 1;
							cnt <= cnt-1;
							S_ready <= 0;
							NO_ERRORS <= 0;
							S1 <= 0;
							S2 <= 0;
							S3 <= 0;
							S4 <= 0;
						end
					end
					
					default: begin
						wait_cnt <= 0;
						cnt <= 0;
						control <= 0;
						A1_IN <= 0;
						B1_IN <= 0;
						S_ready <= 0;
						S1 <= 0;
						S2 <= 0;
						S3 <= 0;
						S4 <= 0;
						NO_ERRORS <= 0;
					end
				endcase
			end
		end

endmodule
