//-----------------------------------------------------------------------------
//
// Title       : Omega_calc
// Design      : decoder_5bit
// Author      : Sindre
// File        : Omega_calc.v
//
//-----------------------------------------------------------------------------
//
// Description : This module calculates the omega(x) coefficients after
//				 error loactor coefficients have been calculated
//
//-----------------------------------------------------------------------------
`timescale 1 ns / 1 ps

//{module {Omega_calc}}
module Omega_calc #(parameter width=5)(
	input clk,			 //clock signal
	input reset,		 //reset signal
	input enable,		 //enable signal
	input L_ready,		 //error locator poly. ready signal
	input [4:0] S1,		 //syndrome value 1
	input [4:0] S2,		 //syndrome value 2
	input [4:0] S3,		 //syndrome value 3
	input [4:0] S4,		 //syndrome value 4
	input [4:0] L0,		 //locator poly. coefficient 0 input
	input [4:0] L1,		 //locator poly. coefficient 1 input
	input [4:0] L2,		 //locator poly. coefficient 2 input
	output reg LW_ready, //error evaluator poly. ready signal
	output [4:0] W0,	 //evaluator poly. coefficient 0 output
	output [4:0] W1		 //evaluator poly. coefficient 1 output
	);
	
	//internal constants
	parameter t = 2;	   		//number of coeff to calculate
	parameter size = 3;
	parameter IDLE = 3'b000; 	//0
	parameter STEP1 = 3'b001;	//1
	parameter STEP2 = 3'b011;	//3
	parameter STEP3 = 3'b010;	//2
	parameter STEP4 = 3'b110;	//6
	
	//internal variables
	reg [width-1:0] S [1:4];	// internal storage for syndrome values
	reg [width-1:0] L [0:2]; 	// register with error locator polynomial coeff
	reg [width-1:0] W [0:1];	// register with error evaluator polynomial coeff
	reg [2:0] N;				//counter, max value 4 
	reg [1:0] i;				//counter, max value 2
	reg n_reset;				//internal reset of register
	
	reg [size-1:0] state;
	reg [size-1:0] next_state;	
	reg gclk;
	reg qint;
	
	//input variables to modules
	reg [width-1:0] M01;		//M<value> indicates an input to a multiplier
	reg [width-1:0] M02;
	//internal wires
	wire [width-1:0] NET01;	   //NET<value> indicates an internal wire
	wire [width-1:0] NET02;
	wire [width-1:0] SUM01;	   //SUM<value> indicates an output from an multiplier or adder
	
	
	//module instantiations
	//modules used for performing finite field addition and multiplication
	Full_mult #(.width(width)) FFM01(
	.A	(M01),
	.B	(M02),
	.Y	(NET01)
	);
	
	Adder #(.width(width)) FFA01(
	.in1	(NET01),
	.in2	(NET02),
	.sum	(SUM01)
	);
   	
	D_reg #(.width(width)) REG01(
	.clk	(clk),
	.reset	(n_reset),
	.enable	(enable),
	.d_in	(SUM01),
	.d_out	(NET02)
	);
	
	assign W0 = W[0];	//error evaluator polynomial coefficient 0
	assign W1 = W[1];	//error evaluator polynomial coefficient 1
	
	//clock gating
	always@(clk or enable)
	begin
		if(clk == 0)begin
			qint <= (!enable);
		end
		
		gclk <= (!qint) && clk;
	end
	
	//FSM
	always @(posedge gclk or negedge reset)
		begin: FSM_SEQ
			if (reset == 0) begin
				state <= IDLE;
			end else if(enable)begin
				state <= next_state;
			end
		end
	
	//change state section
	always @(state or L_ready or N or i)
	begin: FSM_COMBO
		case(state)
			IDLE: begin
				if(L_ready == 1)begin
					next_state = STEP1;
				end else begin
					next_state = IDLE;
				end
			end
			
			STEP1: begin
				next_state = STEP2;
			end
			
			STEP2: begin
				next_state = STEP3;
			end
			
			STEP3: begin
				if (N == t-1)begin
					if(i == N)begin
						//change state when all coeffs have been calculated 
						next_state = STEP4;
					end else begin
						next_state = STEP3;
					end
				end else begin
				if(i == N)begin
						//calculate second coeff when the first one is calculated
						next_state = STEP2;
					end else begin
						next_state = STEP3;
					end
				end		
			end
			
			STEP4: begin
				next_state = IDLE;	
			end
		 
			default: next_state = IDLE;
		endcase	
	end
		
	always @(posedge gclk or negedge reset)
		begin: FSM_OUT
			if(reset == 0)begin
				//reset all registers
				LW_ready <= 0;
				n_reset <= 0;
				M01 <= 0;
				M02 <= 0;
				S[1] <= 0;
				S[2] <= 0;
				S[3] <= 0;
				S[4] <= 0;
				L[0] <= 0;
				L[1] <= 0;
				L[2] <= 0;
				W[0] <= 0;
				W[1] <= 0;
				N <= 0;
				i <= 0;
			end else if(enable) begin
				case(state)
					IDLE: begin
						N <= 0;
						i <= 0;
						LW_ready <= 0;
						n_reset <= 0;
						if(L_ready)begin
							//temp store syndrome values and
							//locator poly coeff while evaluator
							//coeff are calcualted
							S[1] <= S1;
							S[2] <= S2;
							S[3] <= S3;
							S[4] <= S4;
							L[0] <= L0;
							L[1] <= L1;
							L[2] <= L2;
						end else begin
							S[1] <= S1;
							S[2] <= S2;
							S[3] <= S3;
							S[4] <= S4;
							L[0] <= 0;
							L[1] <= 0;
							L[2] <= 0;
						end
					end
					
					STEP1: begin
						//calcualte first partial result
						n_reset <= 0;
						M01 <= S[N+1];
						M02 <= L[0];
						N <= N+1;
					end
					
					STEP2: begin
						//calcualte partial result for coeff
						n_reset <= 1;
						W[N-1] <= SUM01;
						M01 <= S[N-i+1];
						M02 <= L[0];
						i <= i+1;
					end
					
					STEP3: begin
						n_reset <= 1;
						if (N == t-1)begin
							if(i == N)begin
								//calculate last partial result for last coeff
								W[N] <= SUM01;
								M01 <= S[N-i+1];
								M02 <= L[i];
							end else begin
								//calcualte partial result for coeff
								M01 <= S[N-i+1];
								M02 <= L[i];
								W[N] <= SUM01;
								i <= i+1;
							end
						end else begin
							if(i == N)begin
								//calcualte next coeff
								W[N] <= SUM01;
								M01 <= S[N-i+1];
								M02 <= L[i];
								N <= N+1;
								i <= 0;
							end else begin
								//calcualte partial result for coeff
								M01 <= S[N-i+1];
								M02 <= L[i];
								W[N] <= SUM01;
								i <= i+1;
							end
						end
					end
					
					STEP4: begin
						W[N] <= SUM01;
						//error loactor and evaluator polynomials are ready
						LW_ready <= 1; 
					end
					
					default: begin
						LW_ready <= 0;
						n_reset <= 0;
						M01 <= 0;
						M02 <= 0;
						S[1] <= 0;
						S[2] <= 0;
						S[3] <= 0;
						S[4] <= 0;
						L[0] <= 0;
						L[1] <= 0;
						L[2] <= 0;
						W[0] <= 0;
						W[1] <= 0;
						N <= 0;
						i <= 0;	
					end
				endcase
			end
		end

endmodule
